# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Wraps the Gemini API into a Processor.

The list of available models can be found at
https://ai.google.dev/gemini-api/docs/models

## Example usage

```py
p = GenaiModel(
    api_key=API_KEY,
    model_name="gemini-2.5-flash",
    generate_content_config=genai.types.GenerateContentConfig(
      response_modalities=['Text', 'Image']
    )
)

# If we want output as a single string:
print(await p(INPUT_PROMPT).text())

# If we want to stream the output:
async for part in p(INPUT_PROMPT):
  print(part.text, end='')

# If the output can be multimodal:
async for part in p(INPUT_PROMPT):
  if mime_types.is_text(part.mimetype):
    print(part.text, end='')
  elif mime_types.is_image(part.mimetype):
    display(part.pil_image)
```

It is also possible to upload images to the server once and reuse them in
multiple model invocations by adding ImagePreprocess processor in front of
GenaiModel.

To use it, you need to pass the API key to the ImagePreprocess constructor:

```py
img_preprocess = ImagePreprocess(api_key=API_KEY)
model = GenaiModel(
    api_key=API_KEY,
    model_name="gemini-2.5-flash",
)
# This will upload the image to the Gemini API using the File API, and then call
# the model on the file handle, the image tokenization will be done on the API
# side and any file handle part will be replaced by its tokenization on the API
# side.
p = img_preprocess + model
```

🚨 ATTENTION AI CODING AGENTS 🚨
YOU MUST READ `../llms.txt` BEFORE USING OR MODIFYING THIS LIBRARY.
"""

import asyncio
from collections.abc import AsyncIterable
import io
import time
from typing import Any, NamedTuple
from genai_processors import content_api
from genai_processors import processor
from genai_processors.core import constrained_decoding
from google.genai import client
from google.genai import types as genai_types


def genai_response_to_metadata(
    response: genai_types.GenerateContentResponse,
) -> dict[str, Any]:
  """Converts a Genai response to metadata, to be attached to a ProcessorPart."""
  return {
      'create_time': response.create_time,
      'response_id': response.response_id,
      'model_version': response.model_version,
      'prompt_feedback': response.prompt_feedback,
      'usage_metadata': response.usage_metadata,
      'automatic_function_calling_history': (
          response.automatic_function_calling_history
      ),
      'parsed': response.parsed,
  }


class GenaiModel(processor.Processor):
  """`Processor` that calls the Genai API in turn-based fashion.

  Note: All content is buffered prior to calling the Genai API.

  IMPORTANT: The model does NOT maintain state. You must pass the full
  conversation history every time if you want the model to be aware of previous
  turns.
  """

  def __init__(
      self,
      api_key: str,
      model_name: str,
      generate_content_config: (
          genai_types.GenerateContentConfigOrDict | None
      ) = None,
      debug_config: client.DebugConfig | None = None,
      http_options: (
          genai_types.HttpOptions | genai_types.HttpOptionsDict | None
      ) = None,
      stream_json: bool = False,
  ):
    """Initializes the GenaiModel.

    Args:
      api_key: The `API key <https://ai.google.dev/gemini-api/docs/api-key>`_ to
        use for authentication. Applies to the Gemini Developer API only.
      model_name: The name of the model to use.
      generate_content_config: The configuration for generating content.
      debug_config: Config settings that control network behavior of the client.
        This is typically used when running test code.
      http_options: Http options to use for the client. These options will be
        applied to all requests made by the client. Example usage: `client =
        genai.Client(http_options=types.HttpOptions(api_version='v1'))`.
      stream_json: By default, if a `response_schema` is present in the
        `generate_content_config`, this processor will buffer the model's JSON
        output and parse it into dataclass/enum instances. Set this to True to
        disable this behavior and stream the raw JSON text instead.

    Returns:
      A `Processor` that calls the Genai API in turn-based fashion.

    ## Model Name Usage

    Supported formats for **Vertex AI API** include:

      * The Gemini model ID, for example: 'gemini-2.0-flash'
      * The full resource name starts with 'projects/', for example:
      'projects/my-project-id/locations/us-central1/publishers/google/models/gemini-2.0-flash'
      * The partial resource name with 'publishers/', for example:
      'publishers/google/models/gemini-2.0-flash' or
      'publishers/meta/models/llama-3.1-405b-instruct-maas' / separated
      publisher and model name, for example: 'google/gemini-2.0-flash' or
      'meta/llama-3.1-405b-instruct-maas'

    Supported formats for **Gemini API** include:

      * The Gemini model ID, for example: 'gemini-2.0-flash'
      * The model name starts with 'models/', for example:
      'models/gemini-2.0-flash'
      * For tuned models, the model name starts with 'tunedModels/', for
      example: 'tunedModels/1234567890123456789'
    """
    self._client = client.Client(
        api_key=api_key,
        debug_config=debug_config,
        http_options=http_options,
    )
    self._model_name = model_name
    self._generate_content_config = generate_content_config
    self._parser = None

    schema = None
    if generate_content_config:
      if isinstance(generate_content_config, genai_types.GenerateContentConfig):
        if hasattr(generate_content_config, 'response_schema'):
          schema = generate_content_config.response_schema
      elif isinstance(generate_content_config, dict):
        schema = generate_content_config.get('response_schema')

    # If schema is present and the user has not opted for raw JSON streaming,
    # set up the JSON decoding processor.
    if schema and not stream_json:
      self._parser = constrained_decoding.StructuredOutputParser(schema)

  async def _generate_from_api(
      self, content: content_api.ContentStream
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    """Internal method to call the GenAI API and stream results."""
    contents = await content.gather()
    if not contents:
      return

    async for res in await self._client.aio.models.generate_content_stream(
        model=self._model_name,
        contents=content_api.to_genai_contents(contents),
        config=self._generate_content_config,
    ):
      res: genai_types.GenerateContentResponse = res
      if res.candidates:
        content = res.candidates[0].content
        if content and content.parts:
          for part in content.parts:
            yield processor.ProcessorPart(
                part,
                metadata=genai_response_to_metadata(res),
                role=content.role or 'model',
            )

  async def call(
      self, content: content_api.ContentStream
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    api_stream = self._generate_from_api(content)

    if self._parser:
      async for part in self._parser(api_stream):
        yield part
    else:
      async for part in api_stream:
        yield part


class ImagePreprocess(processor.PartProcessor):
  """Processor that prepares images for the GenaiModel."""

  class ImageFile(NamedTuple):
    """Class to represent an image file."""

    file: genai_types.File
    creation_time_sec: float

  def __init__(self, api_key: str, ttl_secs: int | None = None):
    """Initializes the ImagePreprocess.

    Args:
      api_key: The `API key <https://ai.google.dev/gemini-api/docs/api-key>`_ to
        use for authentication. Applies to the Gemini Developer API only.
      ttl_secs: A best-effort TTL in seconds for the uploaded image. Gemini API
        has a server-side non-adjustable 48h TTL. If this is not enough and the
        app exceeds the 20GB limit, this TTL can be used to keep it under
        control.
    """
    self._api_key = api_key
    self._ttl_secs = ttl_secs
    self._client = client.Client(api_key=api_key)
    self._files = asyncio.Queue[ImagePreprocess.ImageFile | None]()
    self._deletion_task = (
        asyncio.create_task(self._delete_images()) if ttl_secs else None
    )

  async def _delete_images(self):
    """Deletes all the images uploaded by the client."""
    while file := await self._files.get():
      await asyncio.sleep(
          max(0, file.creation_time_sec + self._ttl_secs - time.time())
      )
      await asyncio.to_thread(self._client.files.delete, name=file.file.name)

  def match(self, part: content_api.ProcessorPart) -> bool:
    return content_api.is_image(part.mimetype)

  async def call(
      self, part: content_api.ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    image_bytes = io.BytesIO(part.bytes)
    file_part = await asyncio.to_thread(
        self._client.files.upload,
        file=image_bytes,
        config=genai_types.UploadFileConfig(mime_type=part.mimetype),
    )
    await self._files.put(
        ImagePreprocess.ImageFile(file=file_part, creation_time_sec=time.time())
    )
    yield file_part



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Research Agent prompts."""

TOPIC_GENERATION_PREAMBLE = """You are an expert at generating topics for research, based on the user's content.

Your first task is to devise a number of concrete research areas needed to address the user's content.

E.g., if the user's content was 'Starting a vegetable garden in my small backyard in London', you would provide up to {num_topics} topics to be individually researched, such as "what vegetables can grow in the London climate?", and "vegetables that can grow with limited space".

E.g. for "vegetables that can grow with limited space", the 'relationship' could be 'We need to know which vegetables can grow in limited space, given the user wants to start a vegetable garden in their **small** backyard'.
"""


TOPIC_RESEARCH_PREAMBLE = """You are an expert at performing "Deep Research" for users.

Return your detailed research on this topic, in relation to the user's input.

Your research will be collated & provided to another expert, who will combine the research from various topic areas to provide one cohesive answer to the user; so DO NOT exceed the scope of the topic you have been asked to investigate.

DO NOT include citation numbers in your research.
"""

SYNTHESIS_PREAMBLE = """You are an expert at evaluating research results and synthesizing them into a single coherent piece of research.

You will be provided with the user's original content, and the research that has been performed on various topics related to that content.

Please produce a single piece of synthesized research, which collates the provided research into a single coherent piece, which can be used to directly answer the user's original content.

Make sure to reference each of the topics researched at least once in your synthesis.
"""



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for implementing tools."""

from typing import Any, Callable, Collection, Union

import docstring_parser
from google.genai import _transformers
from google.genai import types as genai_types


def raise_for_gemini_server_side_tools(
    tools: list[genai_types.Tool], *, allow_list: Collection[str] = ()
) -> None:
  """Raises ValueError if the tool list contains a server-side tool.

  Gemini API provides a set of server-side tools. In general, they are not
  available on other LLM implementations. This helper method allows warning the
  developer if they try to use this unimplemented functionality.

  Args:
    tools: List of tools.
    allow_list: List of server-side tools to allow.
  """
  for tool in tools:
    for tool_name in (
        'retrieval',
        'google_search',
        'google_search_retrieval',
        'enterprise_web_search',
        'google_maps',
        'url_context',
        'code_execution',
        'computer_use',
    ):
      if getattr(tool, tool_name) is not None and tool_name not in allow_list:
        raise ValueError(f'Tool {tool_name} is not supported.')


def to_schema(
    schema: Union[genai_types.SchemaUnion, genai_types.SchemaUnionDict],
) -> genai_types.Schema:
  """Returns a JSON schema given a Python object representing the schema.

  GenAI SDK accepts a variety of Python objects, such as Enum, dataclass or a
  dictionary of Schema objects as schemas. This provides an easy way to
  constrain the LLM to generate JSON compatible with the needed type. This
  utility is useful for defining the expected structure of data for tool calling
  or constrained decoding with a model.

  Usage:
    json_schema = to_schema(...).json_schema.model_dump(
        mode='json', exclude_unset=True)

  Args:
    schema: The Python object representing the schema to convert. See
      https://ai.google.dev/gemini-api/docs/structured-output for the supported
        types.

  Returns:
    A `genai_types.Schema` object representing the schema.
  """
  return _transformers.t_schema(  # pytype: disable=wrong-arg-types
      _FakeClient(), schema
  )


def to_json_schema(
    schema: Union[genai_types.SchemaUnion, genai_types.SchemaUnionDict, None],
) -> str | None:
  """Returns a JSON schema given a Python object representing the schema.

  GenAI SDK accepts a variety of Python objects, such as Enum, dataclass or a
  dictionary of Schema objects as schemas. This provides an easy way to
  constrain the LLM to generate JSON compatible with the needed type. This
  utility is useful for defining the expected structure of data for tool calling
  or constrained decoding with a model.

  Args:
    schema: The Python object representing the schema to convert. See
      https://ai.google.dev/gemini-api/docs/structured-output for the supported
        types.

  Returns:
    A `genai_types.Schema` object representing the schema or None if the given
    `schema` was None.
  """

  if schema:
    return to_schema(schema).json_schema.model_dump(
        mode='json', exclude_unset=True
    )
  else:
    return None


def to_function_declarations(
    tool_list: list[genai_types.Tool | Callable[..., Any]],
) -> list[genai_types.FunctionDeclaration]:
  """Converts a list of tools to a list of function declarations."""
  tools: list[genai_types.Tool] = []
  for t in tool_list:
    if callable(t):
      fdecl = genai_types.FunctionDeclaration.from_callable_with_api_option(
          callable=t, api_option='GEMINI_API'
      )
      parsed_docstring = docstring_parser.parse(t.__doc__)
      fdecl.description = ''
      if parsed_docstring.short_description:
        fdecl.description += f'{parsed_docstring.short_description}'
      if parsed_docstring.long_description:
        fdecl.description += f'\n\n{parsed_docstring.long_description}'
      if fdecl.parameters:
        for param in parsed_docstring.params:
          if param.arg_name in fdecl.parameters.properties:
            fdecl.parameters.properties[param.arg_name].description = (
                param.description
            )
        tools.append(genai_types.Tool(function_declarations=[fdecl]))
    else:
      tools.append(t)
  raise_for_gemini_server_side_tools(tools)

  function_declarations: list[genai_types.FunctionDeclaration] = []
  for tool in tools:
    function_declarations.extend(tool.function_declarations or ())

  return function_declarations


def function_declaration_to_json(
    fdecl: genai_types.FunctionDeclaration,
) -> dict[str, Any]:
  """Returns a JSON representation of a FunctionDeclaration."""
  return {
      'type': 'function',
      'function': {
          'name': fdecl.name,
          'description': fdecl.description,
          'parameters': to_json_schema(fdecl.parameters),
      },
  }


# TODO(kibergus): Remove this hack once Genai SDK allows None as the client.
class _FakeClient:
  """A fake genai client to invoke t_schema."""

  def __init__(self):
    self.vertexai = False



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Windowing processors for working with long or realtime streams.

When dealing with long streams (e.g. video file) or doing contigous inference on
realtime streams one may need to slice them in-to prompts to be sent to an LLM.
This module contains utilities for that.

Also see genai_processors.core.realtime.LiveProcessor for bidirectional
"Live" chat implementation.
"""

import asyncio
import collections
from collections.abc import AsyncIterable, Awaitable
import contextlib
import time
from typing import Callable

from genai_processors import content_api
from genai_processors import processor
from genai_processors import streams


ProcessorPart = content_api.ProcessorPart
Processor = processor.Processor


class RollingPrompt:
  """Rolling prompt (aka iterator of prompts) for conversation processors.

  NOTE: This is a low level utility. Prefer using realtime.LiveProcessor or
  window.Window which provide higher level abstractions.

  This class acts as a buffer and organizer for a continuous, "infinite" stream
  of multimodal parts (e.g. from a real-time camera feed). It transforms this
  stream into finite, time-segmented prompts suitable for turn-based models.

  Adding Parts:
    add_part(part): Immediately appends a multimodal part to the current prompt.
    stash_part(part): Temporarily stores a part to be appended later. This is
      useful when it's not the client's "turn" to send parts (e.g. while a
      model is generating a response) and you want to delay its inclusion in the
      active prompt.
    apply_stash(): Appends all currently stashed parts to the active prompt.

  Accessing prompts:
    pending() -> `AsyncIterable[ProcessorPart]`: Returns an asynchronous
      iterator representing the current prompt state. This iterator first yields
      the compressed conversation history, then *continues* to stream any new
      parts added via `add_part` or `apply_stash`. This streaming behavior
      allows models to begin processing the current turn *before it is fully
      completed*, significantly minimizing Time To First Token (TTFT).
   finalize_pending(): This method should be called when it is time for the
     underlying model to make the turn.
     * The `AsyncIterable` previously returned by `pending()` will cease
       yielding parts after all current parts have been delivered, signaling the
       model to generate its response.
     * The history compression is applied to accumulated conversation (which
       does not include the ongoing model response, as it has not arrived yet).
     * A new `pending` iterator is then implicitly created for the *next* turn.
     * The compressed history is written to this new iterator, and any parts
       added subsequently will be directed to this new turn's pendig prompt.

  Consider a scenario where a model is actively generating a response, but the
  real-time stream (e.g. a video feed) continues to produce new parts (e.g.
  image frames). Directly using `add_part` for these new frames would
  incorrectly place them *within* or before the ongoing model response, falsely
  suggesting the model considered these new images during its computation for
  the *current* response.

  To address this, such parts coming from the application should be added using
  `stash_part` and appended to the prompt after the model turn is completed
  using `apply_stash`.
  """

  def __init__(
      self,
      *,
      duration_prompt_sec: float | None = None,
      compress_history: (
          Callable[[collections.deque[ProcessorPart]], Awaitable[None]] | None
      ) = None,
  ):
    """Initializes the rolling prompt.

    Args:
      duration_prompt_sec: the length of the prompt in terms of the time when
        the parts were added: we consider any part added after now -
        duration_prompt_sec. Set to None to keep all incoming parts. Set to 10
        minutes by default.
      compress_history: A callable that takes a deque of ProcessorPart and
        modifies it in place to compress the history.
    """
    # Current prompt as a queue.
    self._pending = asyncio.Queue[ProcessorPart | None]()
    # Main prompt content used to build the next prompt and the prompt queues.
    self._conversation_history = collections.deque(maxlen=10_000)
    # stashed parts to be added to the prompt later.
    self._stash: list[ProcessorPart] = []

    async def _noop_compress_history(
        _: collections.deque[ProcessorPart],
    ) -> None:
      pass

    self._compress_history = compress_history or _noop_compress_history
    if duration_prompt_sec:
      if compress_history:
        raise ValueError(
            'compress_history and duration_prompt_sec can not be set'
            ' simultaneously.'
        )

      self._compress_history = drop_old_parts(duration_prompt_sec)

  def add_part(
      self,
      part: ProcessorPart,
  ) -> None:
    """Adds a part to the current prompt."""
    part.metadata.setdefault('capture_time', time.perf_counter())
    self._conversation_history.append(part)
    self._pending.put_nowait(part)

  def stash_part(self, part: ProcessorPart):
    """Stashes the part to be appended to the prompt later using `apply_stash`.

    If it is not our turn to send parts (e.g. model is currently generating),
    we can stash them and append to the prompt when we get the turn (after the
    model generation is done).

    Args:
      part: The part to stash.
    """
    self._stash.append(part)

  def apply_stash(self):
    """Append all parts from the stash to the prompt."""
    for part in self._stash:
      self.add_part(part)
    self._stash = []

  def pending(self) -> content_api.ContentStream:
    """Returns the current pending prompt.

    Note that the same AsyncIterable is returned unless `finalize_pending` is
    called which creates a new pending prompt. So consuming Parts from it will
    affect all callers of `pending`.
    """
    return content_api.ContentStream(
        parts_generator=streams.dequeue(self._pending)
    )

  async def finalize_pending(self) -> None:
    """Close the current pending prompt and starts a new one.

    * The iterator previously returned from `pending` will stop after all Parts
      added so far.
    * Then history compression is applied.
    * At last, a new `pending` iterator is created and the compressed history
      is written to it.
    * Any parts added afterwards will go to the new iterator.
    """
    # Close the current queue.
    self._pending.put_nowait(None)
    # And create a new one.
    self._pending = asyncio.Queue[ProcessorPart | None]()
    await self._compress_history(self._conversation_history)
    for part in self._conversation_history:
      self._pending.put_nowait(part)


def drop_old_parts(
    age_sec: float,
) -> Callable[[collections.deque[ProcessorPart]], Awaitable[None]]:
  """A history compression policy that drops parts older than `age_sec`."""

  async def policy(history: collections.deque[ProcessorPart]):
    time_cut_point = time.perf_counter() - age_sec
    while history and history[0].metadata['capture_time'] < time_cut_point:
      history.popleft()

  return policy


def keep_last_n_turns(
    turns_to_keep: int,
) -> Callable[[collections.deque[ProcessorPart]], Awaitable[None]]:
  """Returns a history compression policy that keeps turns_to_keep last turns.

  Args:
    turns_to_keep: How many turns to keep, including the turn that comes after
      the policy is applied.
  """

  async def policy(history: collections.deque[ProcessorPart]):
    turns = 0
    for part in history:
      if content_api.is_end_of_turn(part):
        turns += 1

    turns_to_drop = turns + 1 - turns_to_keep

    while turns_to_drop > 0:
      if content_api.is_end_of_turn(history[0]):
        turns_to_drop -= 1
      history.popleft()

  return policy


class Window(Processor):
  """Invokes the given processor on a sliding window across incoming content.

  Accumulates incoming content. When `content_api.is_end_of_turn(part)` is
  encountered, `window_processor` is invoked with the accumulated content.

  NOTE: You can put a processor before the Window which will mark Parts with
  `part.metadata['turn_complete'] = true` or filter-out unwanted parts. For
  example when detecting events in a video stream you may leave only image parts
  and mark each as turn_complete.

  The output of `window_processor` is propagated to the Window output, but
  unlike `LiveProcessor` it is not kept in the prompt of `window_processor` and
  is not visible to the consecutive `window_processor` invocations.

  Many instances of `window_processor` may be run concurrently, but their output
  will be yielded in order.

  After `window_processor` is invoked, the accumulated prompt is compressed with
  `compress_history`. E.g. one can keep the last N turns or last M seconds or
  use an LLM to summarize the context. Note that new Parts added after the last
  turn won't be compressed. This allows handling each window in a streaming
  fashion.

  NOTE: If more sophisticated compression is needed (e.g. you want to leave
  image frames with increasing backoff: now, -1s, -5s, -15s, -30s) this can be
  done on per-window basis inside of `window_processor`. Similarly one can use
  Preamble/Postamble processors or set a system prompt to inject additional
  instructions in the `window_processor`.
  """

  def __init__(
      self,
      window_processor: Processor,
      compress_history: (
          Callable[[collections.deque[ProcessorPart]], Awaitable[None]] | None
      ) = None,
      max_concurrency: int = 0,
      stride: int = 1,
  ):
    """Initializes the window processor.

    Args:
      window_processor: The processor to invoke on the window.
      compress_history: A callable that takes a deque of ProcessorPart and
        modifies it in place to compress the history.
      max_concurrency: The maximum number of concurrent window_processor
        invocations. If 0 or less, concurrency is unlimited.
      stride: Only process every `stride` window, skipping the rest. Must be >=
        1.
    """
    if stride < 1:
      raise ValueError('stride must be >= 1')
    self._window_processor = window_processor
    self._compress_history = compress_history
    self._stride = stride
    if max_concurrency > 0:
      self._semaphore = asyncio.Semaphore(max_concurrency)
    else:
      self._semaphore = contextlib.nullcontext()

  async def _consume_content(
      self,
      content: AsyncIterable[ProcessorPart],
      window_results_queue: asyncio.Queue[
          asyncio.Queue[ProcessorPart | None] | None
      ],
      rolling_prompt: RollingPrompt,
  ) -> None:
    """The main loop that creates and manages window processing tasks."""

    window_index = 0

    async def _create_window_task():
      prompt_for_window = rolling_prompt.pending()
      single_window_output_queue = asyncio.Queue[ProcessorPart | None]()
      window_results_queue.put_nowait(single_window_output_queue)

      async def run_window_processor():
        async with self._semaphore:
          try:
            async for part in self._window_processor(prompt_for_window):
              await single_window_output_queue.put(part)
          finally:
            await single_window_output_queue.put(
                content_api.ProcessorPart.end_of_turn()
            )
            await single_window_output_queue.put(None)

      processor.create_task(run_window_processor())

    await _create_window_task()
    window_index += 1

    async for part in content:
      rolling_prompt.add_part(part)
      if content_api.is_end_of_turn(part):
        await rolling_prompt.finalize_pending()
        if window_index % self._stride == 0:
          await _create_window_task()
        window_index += 1

    await rolling_prompt.finalize_pending()

    await window_results_queue.put(None)  # Signal end of all windows.

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:
    """The main entry point for the processor."""
    rolling_prompt = RollingPrompt(compress_history=self._compress_history)
    window_results_queue = asyncio.Queue[
        asyncio.Queue[ProcessorPart | None] | None
    ]()

    consume_content_task = processor.create_task(
        self._consume_content(
            content,
            window_results_queue,
            rolling_prompt,
        )
    )
    while single_window_output_queue := await window_results_queue.get():
      while part := await single_window_output_queue.get():
        yield part
    await consume_content_task



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

r"""Dynamic Widgets demo WebSocket server for AI Studio.

A server / backend to run the Dynamic Widgets demo on AI Studio. Can be used
from this applet:
https://aistudio.google.com/app/apps/github/google-gemini/genai-processors/tree/main/examples/widgets/ais_app

See widgets.py for the actual implementation.

To run the server locally:

 * Install the dependencies with `pip install genai-processors[live]`.
 * Define a GOOGLE_API_KEY environment variable with your API key.
 * Launch the widgets agent: `python3 ./widgets_ais.py --alsologtostderr`.
 * Access the applet at
 https://aistudio.google.com/app/apps/github/google-gemini/genai-processors/tree/main/examples/widgets/ais_app.

A good prompt to test the server is:
  Describe what conic sections are illustrating each of them with a plot.
"""

import asyncio
import os
from typing import Any

from absl import app
from absl import flags
from genai_processors import processor
from genai_processors.examples import live_server
import widgets

_PORT = flags.DEFINE_integer(
    'port',
    8765,
    'Port to run this WebSocket server on.',
)
_TRACE_DIR = flags.DEFINE_string(
    'trace_dir',
    None,
    'If set, enable tracing and write traces to this directory.',
)


def create_dr_widget(
    config: dict[str, Any],
) -> processor.Processor:
  """Creates a dynamic widgets processor."""
  del config  # Unused.
  api_key = os.environ['GOOGLE_API_KEY']
  return widgets.create_dr_widget(
      api_key=api_key,
  )


def main(argv):
  del argv
  asyncio.run(
      live_server.run_server(
          create_dr_widget,
          port=_PORT.value,
          trace_dir=_TRACE_DIR.value,
      )
  )


if __name__ == '__main__':
  app.run(main)



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Handles extracting speech from streaming audio content parts.

Uses Google Cloud Speech API to transcribe audio parts into text parts.

install google cloud speech client with:

```python
pip install --upgrade google-cloud-speech
```

See the `speech_to_text_cli.py` script for a usage example and how to test it
locally. It is recommended to test the quality of the transcription with
different models and recognizers.
"""

import asyncio
from collections.abc import AsyncIterable
import dataclasses
import time

from absl import logging
import dataclasses_json
from genai_processors import content_api
from genai_processors import processor
from genai_processors import streams
from google.cloud import speech_v2
import grpc


_SILENT_AUDIO_DELAY_SECONDS = 1

RecognizeStream = grpc.aio.StreamStreamCall[
    speech_v2.types.StreamingRecognizeRequest,
    speech_v2.types.StreamingRecognizeResponse,
]


DEFAULT_SAMPLE_RATE_HZ = 24000
# streaming_recognize RPC has limit on the duration and has to be restarted
# periodically. Instead of waiting for the deadline we try to restart it at
# the moments when that won't cause hiccups.
STREAMING_HARD_LIMIT_SEC = (
    240  # 4 minutes / restart stream even when user is speaking.
)
STREAMING_LIMIT_SEC = (
    180  # 3 minutes / restart stream when user is not speaking.
)

ProcessorPart = content_api.ProcessorPart

TRANSCRIPTION_SUBSTREAM_NAME = 'input_transcription'
ENDPOINTING_SUBSTREAM_NAME = 'input_endpointing'


@dataclasses_json.dataclass_json
@dataclasses.dataclass(frozen=True)
class StartOfSpeech:
  """Start of speech event."""

  pass


@dataclasses_json.dataclass_json
@dataclasses.dataclass(frozen=True)
class EndOfSpeech:
  """End of speech event."""

  pass


class AddSilentPartMaybe(processor.Processor):
  """Adds silent audio parts if no activity after `silent_part_duration_sec`.

  If the stream is empty after a few seconds, the Speech API will close the
  connection. This processor adds silent audio parts to the output stream to
  keep the connection alive.
  """

  def __init__(
      self,
      silent_part_duration_sec: float = 1,
      sample_rate: int = DEFAULT_SAMPLE_RATE_HZ,
  ):
    self._silent_part_duration_sec = silent_part_duration_sec
    self._sample_rate = sample_rate

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:

    logging.info('Transcriber: _process_audio started.')
    last_streamed_audio_time_sec = time.perf_counter()

    async def _insert_silent_audio() -> AsyncIterable[ProcessorPart]:
      """Sends silent audio to the Speech API to keep the stream alive."""
      nonlocal last_streamed_audio_time_sec

      while True:
        await asyncio.sleep(self._silent_part_duration_sec)
        delta_time_sec = time.perf_counter() - last_streamed_audio_time_sec
        if delta_time_sec > self._silent_part_duration_sec:
          yield ProcessorPart(
              value=b'\0' * round(self._sample_rate * delta_time_sec),
              mimetype=f'audio/l16; rate={self._sample_rate}',
          )
          last_streamed_audio_time_sec = time.perf_counter()

    audio_stream = streams.merge(
        [content, _insert_silent_audio()], stop_on_first=True
    )

    async for part in audio_stream:
      last_streamed_audio_time_sec = time.perf_counter()
      yield part

    logging.info('Transcriber: _process_audio finished.')


class _Transcriber(processor.Processor):
  """Transcribes streaming audio using the Cloud Speech API."""

  def __init__(
      self,
      project_id: str,
      recognition_config: speech_v2.types.RecognitionConfig,
      with_endpointing: bool = True,
      substream_endpointing: str = ENDPOINTING_SUBSTREAM_NAME,
      strict_endpointing: bool = True,
      with_interim_results: bool = True,
      substream_transcription: str = TRANSCRIPTION_SUBSTREAM_NAME,
      passthrough_audio: bool = False,
  ):
    """Transcribes audio parts using the Cloud Speech API.

    Args:
      project_id: The project ID to use for the Speech API.
      recognition_config: The recognition config to use for the Speech API. Set
        it up to adjust the sample rate, languages or the recognition model.
      with_endpointing: Whether to yield endpointing events. Endpointing events
        are text parts with the value set to one of the
        `speech_to_text.SpeechEventType` string enums. The endpointing events
        are yielded in the substream defined by substream_endpointing.
      substream_endpointing: The substream name to use for the endpointing
        events.
      strict_endpointing: Whether to send endpointing events only when interim
        results have been found. This avoids yielding endpointing events when
        the user speech is not recognized (e.g. does not return endpointing for
        noise or laughs or coughing, etc.).
      with_interim_results: Whether to yield interim results. If set to False,
        the processor will only yield the final transcription.
      substream_transcription: The substream name to use for the transcription.
      passthrough_audio: Whether to passthrough the audio parts to the output
        stream. The substream name is set to the default one: ''.
    """
    self._config = speech_v2.types.StreamingRecognitionConfig(
        config=recognition_config,
        streaming_features=speech_v2.types.StreamingRecognitionFeatures(
            interim_results=True,
            enable_voice_activity_events=True,
        ),
    )
    self._sample_rate = (
        self._config.config.explicit_decoding_config.sample_rate_hertz
        or DEFAULT_SAMPLE_RATE_HZ
    )
    self._with_endpointing = with_endpointing
    self._substream_endpointing = substream_endpointing
    self._strict_endpointing = strict_endpointing
    self._with_interim_results = with_interim_results
    self._substream_transcription = substream_transcription
    self._project_id = project_id
    self._passthrough_audio = passthrough_audio

  def _make_setup_request(self) -> speech_v2.types.StreamingRecognizeRequest:
    return speech_v2.types.StreamingRecognizeRequest(
        streaming_config=self._config,
        recognizer=(
            f'projects/{self._project_id}/locations/global/recognizers/_'
        ),
    )

  async def call(
      self,
      content: AsyncIterable[ProcessorPart],
  ) -> AsyncIterable[ProcessorPart]:
    """Transcribes streaming audio using the Cloud Speech API."""

    # The output queue is used to yield the audio parts unchanged in the output
    # stream when self._passthrough_audio is True.
    output_queue = asyncio.Queue[ProcessorPart | None]()

    stream_state: dict[str, bool | float] = {
        'start_time_sec': time.perf_counter(),
        'restart_stream': False,
        'user_speaking': False,
        'stream_is_on': True,
    }

    async def request_stream(
        request_queue: asyncio.Queue[
            speech_v2.types.StreamingRecognizeRequest | None
        ],
    ):
      try:
        request_queue.put_nowait(self._make_setup_request())
        async for part in content:
          if not content_api.is_audio(part.mimetype):
            output_queue.put_nowait(part)
            continue
          if self._passthrough_audio:
            output_queue.put_nowait(part)
          if part.part.inline_data is None:
            continue
          if not part.mimetype.lower().startswith(
              'audio/l16'
          ) or not part.mimetype.lower().endswith(f'rate={self._sample_rate}'):
            raise ValueError(
                f'Unsupported audio mimetype: {part.mimetype}. Expected'
                f' audio/l16;[.*]rate={self._sample_rate}.'
            )
          request_queue.put_nowait(
              speech_v2.types.StreamingRecognizeRequest(
                  audio=part.part.inline_data.data,
              )
          )
          delta_time_sec = time.perf_counter() - stream_state['start_time_sec']
          if (
              (delta_time_sec > STREAMING_LIMIT_SEC)
              and not stream_state['user_speaking']
          ) or (delta_time_sec > STREAMING_HARD_LIMIT_SEC):
            stream_state['restart_stream'] = True
            break
      finally:
        request_queue.put_nowait(None)

    async def send_audio_to_speech_api():
      # Instantiates a client.
      try:
        logging.debug('Transcriber: (re)creating client')
        client = speech_v2.SpeechAsyncClient()
        last_endpointing_event = None
        while stream_state['stream_is_on']:
          request_queue = asyncio.Queue[
              speech_v2.types.StreamingRecognizeRequest | None
          ]()
          populate_request_queue = processor.create_task(
              request_stream(request_queue)
          )
          response_stream = await client.streaming_recognize(
              requests=streams.dequeue(request_queue)
          )
          async for response in response_stream:
            if response == grpc.aio.EOF:
              break
            if (
                response.speech_event_type
                == speech_v2.types.StreamingRecognizeResponse.SpeechEventType.SPEECH_ACTIVITY_BEGIN
            ):
              last_endpointing_event = StartOfSpeech()
              stream_state['user_speaking'] = True
              if self._with_endpointing and not self._strict_endpointing:
                last_endpointing_event = 'SPEECH_ACTIVITY_BEGIN_SENT'
                output_queue.put_nowait(
                    ProcessorPart.from_dataclass(
                        dataclass=StartOfSpeech(),
                        substream_name=self._substream_endpointing,
                    )
                )
            if response.results and response.results[0].alternatives:
              if (
                  isinstance(last_endpointing_event, StartOfSpeech)
                  and self._strict_endpointing
              ):
                # We have not sent the SPEECH_ACTIVITY_BEGIN event yet, we
                # waited for the first transcript to appear.
                last_endpointing_event = 'SPEECH_ACTIVITY_BEGIN_SENT'
                output_queue.put_nowait(
                    ProcessorPart.from_dataclass(
                        dataclass=StartOfSpeech(),
                        substream_name=self._substream_endpointing,
                    )
                )
              if text := response.results[0].alternatives[0].transcript:
                metadata = {
                    'is_final': response.results[0].is_final,
                }
                if self._with_interim_results or response.results[0].is_final:
                  output_queue.put_nowait(
                      ProcessorPart(
                          text,
                          role='user',
                          metadata=metadata,
                          substream_name=self._substream_transcription,
                      )
                  )
            if (
                response.speech_event_type
                == speech_v2.types.StreamingRecognizeResponse.SpeechEventType.SPEECH_ACTIVITY_END
            ):
              stream_state['user_speaking'] = False
              if (
                  self._with_endpointing
                  and last_endpointing_event == 'SPEECH_ACTIVITY_BEGIN_SENT'
              ):
                output_queue.put_nowait(
                    ProcessorPart.from_dataclass(
                        dataclass=EndOfSpeech(),
                        substream_name=self._substream_endpointing,
                    )
                )
              last_endpointing_event = None
          if stream_state['restart_stream']:
            stream_state['restart_stream'] = False
            stream_state['stream_is_on'] = True
            stream_state['start_time_sec'] = time.perf_counter()
            client = speech_v2.SpeechAsyncClient()
            populate_request_queue.cancel()
          else:
            stream_state['stream_is_on'] = False
      finally:
        output_queue.put_nowait(None)

    send_task = processor.create_task(send_audio_to_speech_api())
    while part := await output_queue.get():
      yield part
    await send_task


class SpeechToText(processor.Processor):
  """Converts audio parts into text parts."""

  def __init__(
      self,
      project_id: str,
      recognition_config: speech_v2.types.RecognitionConfig | None = None,
      audio_passthrough: bool = False,
      with_endpointing: bool = True,
      substream_endpointing: str = ENDPOINTING_SUBSTREAM_NAME,
      strict_endpointing: bool = True,
      with_interim_results: bool = True,
      substream_transcription: str = TRANSCRIPTION_SUBSTREAM_NAME,
      maintain_connection_active_with_silent_audio: bool = False,
  ):
    """Initializes the SpeechToText processor.

    The speech processor uses the Cloud Speech API to transcribe audio parts
    into text parts. It injects silent audio parts to keep the stream alive
    when the user is not speaking and restarts the connection automatically
    after 3-4 minutes to avoid the stream being closed by the server.

    The processor yields endpointing events when the user starts and stops
    speaking. If with_endpointing is False, the endpointing events are not
    yielded. The endpointing events are yielded in the substream defined by
    substream_endpointing. When strict_endpointing is True, the endpointing
    events are yielded only when interim results have been found. This avoids
    yielding endpointing events when the user speech is not recognized (e.g.
     short noise or sound).

    Args:
      project_id: The project ID to use for the Speech API.
      recognition_config: The recognition config to use for the Speech API. Set
        it up to adjust the sample rate, languages or the recognition model.
      audio_passthrough: Whether to passthrough the audio parts to the output
        stream. The substream name is set to the default one: ''.
      with_endpointing: Whether to yield endpointing events. Endpointing events
        are text parts with the value set to one of the
        `speech_to_text.SpeechEventType` string enums. The endpointing events
        are yielded in the substream defined by substream_endpointing.
      substream_endpointing: The substream name to use for the endpointing
        events.
      strict_endpointing: Whether to send endpointing events only when interim
        results have been found. This avoids yielding endpointing events when
        the user speech is not recognized (e.g. does not return endpointing for
        noise or laughs or coughing, etc.).
      with_interim_results: Whether to yield interim results. If set to False,
        the processor will only yield the final transcription.
      substream_transcription: The substream name to use for the transcription.
      maintain_connection_active_with_silent_audio: Whether to maintain the
        connection active with silent audio. If set to True, the processor will
        inject silent audio parts to keep the stream alive when the processor
        does not receive any audio part. This can be needed if the Speech API
        closes the stream when it does not receive any audio for a long time.
    """
    recognition_config = recognition_config or speech_v2.types.RecognitionConfig(
        explicit_decoding_config=speech_v2.types.ExplicitDecodingConfig(
            sample_rate_hertz=DEFAULT_SAMPLE_RATE_HZ,
            encoding=speech_v2.types.ExplicitDecodingConfig.AudioEncoding.LINEAR16,
            audio_channel_count=1,
        ),
        language_codes=['en-US'],
        model='latest_long',
    )
    self._processor = _Transcriber(
        project_id=project_id,
        recognition_config=recognition_config,
        with_endpointing=with_endpointing,
        substream_endpointing=substream_endpointing,
        strict_endpointing=strict_endpointing,
        with_interim_results=with_interim_results,
        substream_transcription=substream_transcription,
        passthrough_audio=audio_passthrough,
    )
    if maintain_connection_active_with_silent_audio:
      sample_rate = (
          recognition_config.explicit_decoding_config.sample_rate_hertz
          or DEFAULT_SAMPLE_RATE_HZ
      )
      self._processor = (
          AddSilentPartMaybe(
              silent_part_duration_sec=_SILENT_AUDIO_DELAY_SECONDS,
              sample_rate=sample_rate,
          )
          + self._processor
      )

  async def call(
      self,
      content: AsyncIterable[ProcessorPart],
  ) -> AsyncIterable[ProcessorPart]:
    async for part in self._processor(content):
      yield part



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Research agent example, using Processors library.

This agent takes user content as input, generates a set of topics based on
that content, researches each topic, and then synthesizes the research into a
final output.

The agent is composed of three main processors:
  - TopicGenerator: Generates `Topics` based on the user content.
  - TopicResearcher: Researches each `Topic`.
  - TopicVerbalizer: Converts a `Topic` ProcessorPart to a human-readable
  string.
  - The output of this flow is then passed to a GenAI model, which synthesizes
  the research into a final output.

The agent orchestrates these processors, passing the output of one to the
input of the next.
"""

from typing import AsyncIterable

from genai_processors import processor
from genai_processors.core import genai_model
from genai_processors.core import jinja_template
from genai_processors.core import preamble
from google.genai import types

from . import interfaces
from . import prompts
from .processors import topic_generator
from .processors import topic_researcher

ProcessorPart = processor.ProcessorPart


class ResearchAgent(processor.Processor):
  """Research agent example, using Processors library.

  This agent takes user content as input, generates a set of topics based on
  that content, researches each topic, and then synthesizes the research into a
  final output.

  The agent is composed of three main processors:
    - TopicGenerator: Generates `Topics` based on the user content.
    - TopicResearcher: Researches each `Topic`.
    - TopicVerbalizer: Converts a `Topic` ProcessorPart to a human-readable
    string.
    - The output of this flow is then passed to a GenAI model, which synthesizes
    the research into a final output.

  The agent orchestrates these processors, passing the output of one to the
  input of the next.
  """

  def __init__(
      self, api_key: str, config: interfaces.Config = interfaces.Config()
  ) -> None:
    """Initializes the Research Agent.

    Args:
      api_key: The API key to use for the GenAI API.
      config: The configuration for the Research Agent.
    """
    self._config = config

    p_topic_generator = topic_generator.TopicGenerator(
        api_key=api_key,
        config=config,
    )
    p_topic_researcher = topic_researcher.TopicResearcher(
        api_key=api_key,
        config=config,
    )
    p_topic_verbalizer = jinja_template.RenderDataClass(
        template_str=(
            '## {{ data.topic }}\n'
            '*{{ data.relationship_to_user_content }}*'
            '{% if data.research_text|trim != "" %}'
            '\n\n### Research\n\n{{ data.research_text }}'
            '{% endif %}'
        ),
        data_class=interfaces.Topic,
    )
    p_genai_model = genai_model.GenaiModel(
        api_key=api_key,
        model_name=self._config.research_synthesizer_model_name,
        http_options=types.HttpOptions(
            retry_options=types.HttpRetryOptions(
                attempts=100,
            ),
        ),
    )
    p_preamble = preamble.Preamble(
        content=[
            ProcessorPart(prompts.SYNTHESIS_PREAMBLE),
            ProcessorPart('Research text: '),
        ]
    )
    p_suffix = preamble.Suffix(
        content=[
            ProcessorPart('Your synthesized research: '),
        ]
    )
    self._pipeline = (
        p_topic_generator
        + p_topic_researcher
        + p_topic_verbalizer
        + p_preamble
        + p_suffix
        + p_genai_model
    )

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:
    async for content_part in self._pipeline(content):
      yield content_part
    yield processor.status('Produced research synthesis!')



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Command line interface for PDF processor.

Usage:
  python3 pdf_cli.py <pdf_file>

Requires a PDF file to be passed as argument.

This will print the parts extracted by pdf.PDFExtract to stdout.
"""
import argparse
import asyncio
import time

from genai_processors import content_api
from genai_processors.core import pdf


async def run_pdf(content: bytes, filename: str) -> None:
  """Extracts PDF content."""
  p = pdf.PDFExtract()
  metadata = {'original_file_name': filename}
  async for part in p(
      content_api.ProcessorPart(
          content, mimetype=pdf.PDF_MIMETYPE, metadata=metadata
      )
  ):
    print(f'{time.perf_counter()} - {part}')


if __name__ == '__main__':
  parser = argparse.ArgumentParser()
  parser.add_argument('pdf_file')
  args = parser.parse_args()
  pdf_file = args.pdf_file
  with open(pdf_file, 'rb') as f:
    pdf_bytes = f.read()
    print(
        f'{time.perf_counter()} - PDF Processor start extracting PDF:'
        f' {pdf_file}'
    )
    asyncio.run(run_pdf(content=pdf_bytes, filename=pdf_file))



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processor adding timestamps to content."""

import functools
import time
from typing import AsyncIterable

from genai_processors import content_api
from genai_processors import processor


def to_timestamp(seconds: float | int, with_ms: bool = False) -> str:
  """Returns a timestamp as a string to indicate the time of an event."""
  if with_ms:
    seconds = round(seconds * 1000) / 1000
  else:
    seconds = round(seconds)
  minutes = seconds // 60
  seconds = seconds % 60
  if with_ms:
    return f'{minutes:02.0f}:{seconds:06.3f}'
  else:
    return f'{minutes:02.0f}:{seconds:02.0f}'


async def _add_timestamps(
    content: AsyncIterable[content_api.ProcessorPart],
    with_ms: bool = False,
    substream_name: str | None = None,
) -> AsyncIterable[content_api.ProcessorPart]:
  """Adds timestamps to image chunks (to be used by default when streaming)."""
  start = time.perf_counter()
  async for part in content:
    if content_api.is_image(part.mimetype):
      yield content_api.ProcessorPart(
          to_timestamp(time.perf_counter() - start, with_ms=with_ms),
          substream_name=substream_name,
          # Do not trigger a model generate call when the timestamp is added.
          metadata={'turn_complete': False},
      )
    yield part


def add_timestamps(
    with_ms: bool = False,
    substream_name: str | None = None,
) -> processor.Processor:
  """Adds timestamps to image chunks.

  By default the timestamps are added with the format `mm:ss` where
  `mm` is the number of minutes, `ss` is the number of seconds.

  Args:
    with_ms: Whether to add milliseconds to the timestamp. When `True`, the
      timestamp is added with the format `mm:ss.SSS` where `SSS` is the number
      of milliseconds.
    substream_name: The substream name to use for the timestamps.

  Returns:
    A processor that adds timestamps after each image chunk.
  """
  if substream_name is None:
    substream_name = ''
  return processor.processor_function(
      functools.partial(
          _add_timestamps,
          with_ms=with_ms,
          substream_name=substream_name,
      )
  )



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processors operating on text and regular expressions."""

import asyncio
from collections.abc import AsyncIterable, Callable
import dataclasses
import re
import threading
from typing import Literal, Mapping, Type

import bs4
import dataclasses_json
from genai_processors import content_api
from genai_processors import mime_types
from genai_processors import processor
import termcolor


_MAX_LOOP_COUNT = 1000


class MatchProcessor(processor.Processor):
  r"""Processor finding text patterns and extracting them from the input stream.

  There are two modes controlled by the `remove_from_input_stream` parameter.

  1. [True] When the pattern is found, the matched text is removed from the
     input stream and returned as a single part with the substream name
     `substream_output` and the mimetype text/plain.
  2. [False] When the pattern is found, it is not removed from the input stream.
     It is still returned as a single part with the substream name
     `substream_output` and the mimetype text/plain.

  Example [mode 1]:
  ```python
    p = MatchProcessor(
          word_start='[',
          pattern=r'\[.*\]',
          substream_output='from_regex'
        )
    output = processor.apply_sync(p, ['a', 'b', 'c[d', 'e]f','g'])
    ```
    output will contain 6 parts:

    -  content_api.ProcessorPart('a', mimetype='text/plain')
    -  content_api.ProcessorPart('b', mimetype='text/plain')
    -  content_api.ProcessorPart('c', mimetype='text/plain')
    -  content_api.ProcessorPart('[de]',
         mimetype='text/plain',
         substream_name='from_regex',
       )
    -  content_api.ProcessorPart('f', mimetype='text/plain')
    -  content_api.ProcessorPart('g', mimetype='text/plain')

    The text part of `output` will be `abcfg`.

  In [mode 2], the output will be:

    -  content_api.ProcessorPart('a', mimetype='text/plain')
    -  content_api.ProcessorPart('b', mimetype='text/plain')
    -  content_api.ProcessorPart('c[d', mimetype='text/plain')
    -  content_api.ProcessorPart('e]f', mimetype='text/plain')
    -  content_api.ProcessorPart('[de]',
         mimetype='text/plain',
         substream_name='from_regex',
       )
    -  content_api.ProcessorPart('g', mimetype='text/plain')

  When using this processor in a real-time setting, it is recommended to set
  `remove_from_input_stream` to False. This will allow the processor to output
  the parts as soon as possible and not block the stream. If you still need
  to remove the matched text from the input stream, it is advised to define a
  short `word_start` and/or flush conditions (via `flush_fn`) that are met often
  to reset the buffer and to keep producing parts in the output stream.
  """

  def __init__(
      self,
      *,
      pattern: str | re.Pattern[str],
      word_start: str | None = None,
      substream_input: str = '',
      substream_output: str = '',
      flush_fn: Callable[[content_api.ProcessorPart], bool] | None = (None),
      remove_from_input_stream: bool = True,
      transform: (
          Callable[[content_api.ProcessorPart], content_api.ProcessorPart]
          | None
      ) = None,
  ):
    """Extracts text parts from the input stream that match the pattern.

    Only considers text parts from the input stream `substream_input`.

    See class docstring for more details.

    This processor buffers input parts until one of the following happens:

    1. the `pattern` is found, then it outputs all parts in the buffer up to the
      end of the match.
    2. when the `flush_fn` returns True, then it outputs all parts in the
      buffer.

    When `word_start` is set, the following happens:

    3. the word_start is found, then it outputs all the parts before the
      `word_start`.
    4. `word_start` is not found in buffer_text[-len(word_start):], where
       buffer_text is the concatenation of all the text parts in the buffer. All
       parts whose text is in buffer_text[-len(word_start):] are returned.

    While in the buffer, the parts are not output which means they can lead to
    delays in the output stream.

    To avoid such delays, set the `remove_from_input_stream` to False and/or
    define a `flush_fn` that returns True often to discard the parts in the
    buffer frequently.

    Args:
      pattern: pattern to match a text to extract into a part. When
        `remove_from_input_stream` is True, the matched text will be removed
        from the stream and will be replaced by a single extracted part. The
        parts before and after this match will be returned as is. Note that by
        default, re.DOTALL is used to match newlines. To override this behavior,
        pass a re.Pattern object instead.
      word_start: text to match the start of the text that needs to be captured.
        `word_start` is not a regular expression but a plain string that will be
        matched exactly. `word_start` should be a substring of the pattern and
        should indicate that the pattern is about to be matched. Whenever
        `word_start` is found, the parts after it will be buffered (not
        returned) until the pattern is found, the `flush_fn` returns True, or
        the input stream is exhausted. When set to None (default), this logic is
        not applied.
      substream_input: name of the substream to use for the input part.
      substream_output: name of the substream to use for the extracted part.
      flush_fn: function to check when to reset the buffer and yield all the
        parts in the buffer. The part where `flush_fn` returns True will be
        returned as is and will not be matched against the pattern.
      remove_from_input_stream: if True, the processor will remove the matched
        parts from the input stream. If False, the input stream will be
        preserved and the parts will be returned as is quickly. The processor
        will output into its `substream_output` substream once a match is found.
      transform: A transformation to be applied to the matched Parts.
    """
    self._word_start = word_start
    if isinstance(pattern, str):
      self._pattern = re.compile(pattern, re.DOTALL)
    else:
      self._pattern = pattern
    self._substream_input = substream_input
    self._substream_output = substream_output
    self._flush_fn = flush_fn or content_api.is_end_of_turn
    self._remove_from_input_stream = remove_from_input_stream
    if transform:
      self._transform = transform
    else:
      self._transform = lambda part: part

  def _extract_part(
      self, text_buffer: str, part_buffer: list[content_api.ProcessorPart]
  ) -> tuple[list[content_api.ProcessorPart], list[content_api.ProcessorPart]]:
    """Returns the list of parts to yield and the remaining parts to process."""
    to_yield = []
    to_process = part_buffer
    left_over = []
    if (match := self._pattern.search(text_buffer)) is None:
      return to_yield, to_process
    # We have found the pattern, we can yield all the parts up to the
    # beginning of the match and then yield the parts after the match.
    offset = 0
    part_idx = -1
    for c in part_buffer:
      part_idx += 1
      if (
          not content_api.is_text(c.mimetype)
          or c.substream_name != self._substream_input
      ):
        if self._remove_from_input_stream:
          to_yield.append(c)
        continue
      # Find the start of the part to extract. Yields all parts until
      # this start.
      if (offset + len(c.text)) <= match.start():
        part_end = len(c.text)
      else:
        part_end = match.start() - offset
      if part_end > 0 and self._remove_from_input_stream:
        to_yield.append(
            content_api.ProcessorPart(
                c.text[:part_end],
                metadata=c.metadata,
                substream_name=c.substream_name,
                mimetype=c.mimetype,
            )
        )
      if match.start() < offset + len(c.text) and match.start() >= offset:
        to_yield.append(
            self._transform(
                content_api.ProcessorPart(
                    match.group(0),
                    metadata=c.metadata,
                    substream_name=self._substream_output,
                    mimetype=c.mimetype,
                )
            )
        )

      # Find the start of the parts after the match.
      part_start = match.end() - offset
      if part_start < len(c.text) and self._remove_from_input_stream:
        # We have reached the end of the match, there can be another one later
        # in the buffer. We keep the part after the match and stop.
        left_over = [
            content_api.ProcessorPart(
                c.text[part_start:],
                metadata=c.metadata,
                substream_name=c.substream_name,
                mimetype=c.mimetype,
            )
        ]
        break
      offset += len(c.text)
    to_process = left_over + part_buffer[part_idx + 1 :]
    return to_yield, to_process

  async def call(
      self, content: AsyncIterable[content_api.ProcessorPart]
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    part_buffer = []
    async for part in content:
      if not self._remove_from_input_stream:
        yield part
      part_buffer.append(part)
      if self._flush_fn(part):
        if self._remove_from_input_stream:
          for part_b in part_buffer:
            yield part_b
        part_buffer = []
      text_buffer = content_api.as_text(
          part_buffer, substream_name=self._substream_input
      )
      # If the word_start is not in the buffer, we can already yield all the
      # parts up to the last ones that could contain word_start.
      # This is a quick check to avoid buffering more than necessary.
      # Only applies when word_start is set.
      if self._word_start is not None and self._word_start not in text_buffer:
        offset = 0
        idx = 0  # index of first part not yielded.
        for c in part_buffer:
          if (
              not content_api.is_text(c.mimetype)
              and self._remove_from_input_stream
          ):
            yield c
            idx += 1
            continue
          offset += len(content_api.as_text(c))
          if (
              offset < len(text_buffer) - len(self._word_start)
              and self._remove_from_input_stream
          ):
            yield c
            idx += 1
          else:
            break
        part_buffer = part_buffer[idx:]
      else:
        to_yield, part_buffer = self._extract_part(text_buffer, part_buffer)
        for part in to_yield:
          yield part

    # Process the last part which can contain the pattern many times.
    loop_count = 0
    while part_buffer and loop_count < _MAX_LOOP_COUNT:
      text_buffer = content_api.as_text(
          part_buffer, substream_name=self._substream_input
      )
      to_yield, part_buffer = self._extract_part(text_buffer, part_buffer)
      if not to_yield:
        # No match found, we can yield all the parts.
        if self._remove_from_input_stream:
          for part in part_buffer:
            yield part
        break
      else:
        for part in to_yield:
          yield part
    if loop_count >= _MAX_LOOP_COUNT:
      raise RuntimeError(
          'Max loop count reached, the pattern or the input stream is probably'
          ' malformed.'
      )


@dataclasses_json.dataclass_json
@dataclasses.dataclass(frozen=True)
class FetchRequest:
  """Dataclass to represent a request to fetch a document by a URL."""

  url: str


class UrlExtractor(MatchProcessor):
  """Replaces encountered text URLs with strongly typed Parts.

  Also see core/web.py for the infrastructure to fetch the extracted URLs.
  Alternatively see core/github.py for a processor that provides a tailored
  handling for github URLs.

  In some scenarios it is useful to replace URLs mentioned in the prompt with
  the content they point to. In many cases it can be handled by tool calls, but
  if we want to avoid additional roundtrip or the underlying model does not
  support tools, hardwired logic might be preferrable.

  We recommend splitting detecting the URLs and fetching them into separate
  processors. The processor that does the fetching should act on Parts with a
  special MIME type to avoid fetching them unintentionally. And a separate
  processor should decide which URLs should be processed.

  Usage:
    In the basic case use UrlExtractor() without arguments and handle
    FetchRequest parts produced. Alternatively you can define separate
    dataclasses for specific URL prefixes:

      @dataclasses_json.dataclass_json
      @dataclasses.dataclass(frozen=True)
      class YouTubeUrl:
        url: str

    And then tell UrlExtractor to extract them:

      UrlExtractor({
          'https://youtube.': YouTubeUrl,
          'https://github.com': GithubUrl
      })

    Note that all URLs must have the same scheme to allow efficient matching.
  """

  def __init__(
      self,
      urls: Mapping[str, Type] | None = None,  # pylint: disable=g-bare-generic
      *,
      substream_input: str = '',
      substream_output: str = '',
  ):
    """Initiallizes the extractor.

    Args:
      urls: A map from URL prefix (e.g. 'https://github.com') to a Dataclass)
      substream_input: name of the substream to use for the input part.
      substream_output: name of the substream to use for the extracted part.
    """
    if urls is None:
      urls = {'http://': FetchRequest, 'https://': FetchRequest}

    schemes = set()
    for prefix in urls.keys():
      scheme = prefix.split(':')[0]
      if scheme == 'https':
        # Allow mixing http and https URLs.
        # This works because http is a prefix of https.
        scheme = 'http'
      schemes.add(scheme)

    if len(schemes) != 1:
      raise ValueError(
          'All URL prefixes must have the same scheme e.g. https. Got'
          f' {schemes!r}'
      )

    def transform(part: content_api.ProcessorPart):
      for prefix, dataclass in urls.items():
        if part.text.startswith(prefix):
          return content_api.ProcessorPart.from_dataclass(
              dataclass=dataclass(part.text),
              metadata=part.metadata,
              substream_name=part.substream_name,
          )

    super().__init__(
        pattern=re.compile(
            '('
            + '|'.join(urls.keys())
            + r')([^\s<>"\'\u200B]*[^\s<>"\'\u200B\.\,])?',
            re.IGNORECASE,
        ),
        word_start=next(iter(schemes)),
        substream_input=substream_input,
        substream_output=substream_output,
        remove_from_input_stream=True,
        transform=transform,
    )


_TAG_DENYLIST = [
    'head',
    'script',
    'noscript',
    'style',
    'footer',
    'aside',
    'nav',
    'svg',
]
_ATTR_ALLOWLIST = [
    'alt',
    'href',
    'aria-label',
    'aria-level',
    'aria-roledescription',
]


def _clean_html(raw_html: str) -> bs4.BeautifulSoup:
  """Cleans raw HTML into cleaner HTML and plaintext.

  Args:
    raw_html: raw html document - can include a long prefix, e.g. containing the
      response header of some service: all the content is stripped until the
      first occurrence of `<html`.

  Returns:
    a tuple where the first element is the html content, and the second one is
    the page content stripped from all html tags and reformatted.
  """
  # Should catch <html> with and without attribute, e.g. <html locale='..'>.
  # The document could be prefixed with some non-html header.
  html_tag_index = raw_html.find('<html')
  if html_tag_index >= 0:
    html_no_header = raw_html[html_tag_index:]
  else:
    html_no_header = raw_html
  soup = bs4.BeautifulSoup(html_no_header, 'html.parser')
  for entry in soup(_TAG_DENYLIST):
    entry.decompose()
  for tag in soup.descendants:
    if isinstance(tag, bs4.element.Tag):
      tag.attrs = {
          key: value
          for key, value in tag.attrs.items()
          if key in _ATTR_ALLOWLIST
      }
  return soup


class HtmlCleaner(processor.PartProcessor):
  """PartProcessor cleaning up HTML content.

  This part processor will return a new version of a processor part that is of
  the `text/html` MIME type.

  Based on the `cleaning_mode` argument, the content will be cleaned
  accordingly and returned in a new part, that is, many tags and attributes will
  be stripped away. When `plain` mode is selected, the content will be returned
  as plain text with no formatting.

  Note that the cleaning is done per part. It is recommended to collect the
  whole html content into a single part. Otherwise, the cleaning might only be
  done partially, i.e. there is no guarantee that the html content within a
  single part is a valid html text.
  """

  def __init__(self, *, cleaning_mode: Literal['html', 'plain'] = 'plain'):
    self._cleaning_mode = cleaning_mode

  def match(self, part: content_api.ProcessorPart) -> bool:
    return mime_types.is_html(part.mimetype)

  async def call(
      self, part: content_api.ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    match self._cleaning_mode:
      case 'html':
        yield content_api.ProcessorPart(
            _clean_html(part.text).prettify().strip(),
            mimetype=mime_types.TEXT_HTML,
        )
      case 'plain':
        yield _clean_html(part.text).get_text().strip()
      case _:
        raise ValueError(f'Unsupported cleaning mode: {self._cleaning_mode}')


@processor.source()
async def terminal_input(
    prompt: str = '',
) -> AsyncIterable[content_api.ProcessorPartTypes]:
  """Yields lines from the terminal, exits on ctrl+D.

  Note: This only works with Live processors (bidi). Turn-based models will hang
  because this stream does not end until ctrl+D is pressed.

  Args:
    prompt: The prompt to print before reading the input.

  Yields:
    Every line that the user inputs, followed by an end_of_turn part.
  """

  while True:
    try:
      loop = asyncio.get_running_loop()
      future = loop.create_future()

      # We do this hard way instead of a usual asyncio.to_thread() to properly
      # handle cancellations.
      def _read_input():
        try:
          result = input(prompt)
          if not future.cancelled():
            loop.call_soon_threadsafe(future.set_result, result)
        except BaseException as e:
          if not future.cancelled():
            loop.call_soon_threadsafe(future.set_exception, e)

      thread = threading.Thread(target=_read_input, daemon=True)
      thread.start()

      yield await future
      yield content_api.ProcessorPart.end_of_turn()
    except EOFError:
      # Exit on ctrl+D.
      return


async def terminal_output(
    content: AsyncIterable[content_api.ProcessorPartTypes],
    prompt: str = '',
) -> None:
  """Prints the part to the terminal.

  Consumes all the content and prints it to the terminal. Prints the prompt
  when an `end_of_turn` part is encountered.

  The parts are printed with their role in green, red or yellow. The text parts
  are printed in bold. Function calls and responses are printed in cyan.

  Args:
    content: The content to print.
    prompt: The prompt to print when the model is done.
  """
  old_part_role = None
  async for part in content:
    match part.role:
      case 'user':
        color = 'green'
      case 'model':
        color = 'red'
      case _:
        color = 'yellow'
    part_role = part.role or 'default'
    if content_api.is_text(part.mimetype):
      content_text = part.text
    elif part.function_call:
      content_text = (
          f'\n\033[96m FC [{part.function_call.name}]:'
          f' {part.function_call.args}'
      )
    elif part.function_response:
      content_text = (
          f'\n\033[96m FR[{part.function_response.name}]:'
          f' {part.function_response.response}'
      )
    else:
      content_text = f'<{part.mimetype}>'
    if part_role != old_part_role:
      old_part_role = part_role
      print(
          termcolor.colored(
              f'\n{part_role}: {content_text}', color, attrs=['bold']
          ),
          end='',
          flush=True,
      )
    else:
      print(
          termcolor.colored(f'{content_text}', color, attrs=['bold']),
          end='',
          flush=True,
      )
    # Reprint the prompt when the model is done.
    if prompt and content_api.is_end_of_turn(part) and part_role == 'model':
      print('\n' + prompt, end='', flush=True)



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Processor that produces richly formatted responses using widgets.

This is the implementation, use widgets_ais.py to run the server.

This demo demonstrates how to enrich model output with UI elements (widgets)
using async tools. Key concepts include:

*   **Async Tools**: Widgets (e.g., ImageGenerator, PlotGenerator) are
    registered as tools that do not block the model.
*   **Tiered Rendering**:
    * Model decides the widget/content.
    * Tool renders to HTML/low-level representation.
    * UI displays the content.
*   **Direct Streaming**: Tools stream responses directly to the UI using the
    reserved `ui` substream, bypassing the model to avoid context pollution
    and head-of-line blocking.

See README.md for more details.
"""

from collections.abc import AsyncIterable

from genai_processors import content_api
from genai_processors import processor
from genai_processors.core import function_calling
from genai_processors.core import genai_model
from genai_processors.core import realtime
from google.genai import types as genai_types

MODEL = 'gemini-3-flash-preview'
IMAGE_MODEL = 'gemini-2.5-flash-image'


# System Instruction for the root processor.
SYSTEM_INSTRUCTION = [
    'You are a Scientist agent. You help students by explaining how things'
    ' work.'
]


class ImageGenerator:
  """Image generator from a description."""

  def __init__(self, api_key: str):
    self._api_key = api_key
    self._image_generator = genai_model.GenaiModel(
        api_key=self._api_key,
        model_name=IMAGE_MODEL,
    )

  async def create_image_from_description(
      self,
      description: str,
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    """Creates an image from a description.

    The description should be detailed enough to give all the information
    needed to create the image, including style and tone, shapes, shades, etc.
    The image will be visible to the user only, but not you.

    Args:
      description: The description of the image to create.

    Yields:
      The image parts.
    """

    # Notify the model that all is done. Actual output goes directly to the UI.
    # NOTE: we should make this a property of the tool.
    yield content_api.ProcessorPart.from_function_response(
        response='Image generated', will_continue=False
    )

    content = [
        (
            'Generate an illustration based on the following description. '
            'You may produce a short clarification after the image.\n'
        ),
        description,
    ]

    async for part in self._image_generator(content):
      # We use ui substream to send it directly to the UI.
      part.substream_name = 'ui'
      yield part


class PlotGenerator:
  """Plot generator from a description."""

  def __init__(self, api_key: str):
    self._api_key = api_key
    self._plot_generator = genai_model.GenaiModel(
        api_key=self._api_key,
        model_name=MODEL,
    )

  async def create_plot_from_description(
      self,
      description: str,
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    """Creates a plot from a description.

    The description should be detailed enough to give all the information
    needed to create the plot. The plot will be visible to the user only,
    but not you.

    Args:
      description: The description of the plot to create.

    Yields:
      The plot HTML.
    """

    # Notify the model that all is done. Actual output goes directly to the UI.
    yield content_api.ProcessorPart.from_function_response(
        response='Plot generated', will_continue=False
    )

    content = [
        (
            'Generate a standalone HTML file containing an SVG visualization'
            ' of the data described below. Use inline SVG code directly'
            ' embedded in the HTML. Do NOT use any external JavaScript'
            ' libraries (no Bokeh, Plotly, D3, Chart.js, etc.). Output ONLY'
            ' the raw HTML code. Do NOT output markdown code blocks. Do NOT'
            ' output any explanation.\n'
            '\n'
            'The SVG should be self-contained and styled appropriately.'
            ' Include proper axes, labels, and legends as needed. Page height'
            ' must match the SVG height and SVG should take up all the width.'
            ' Do not leave big margins on any side of the SVG.\n'
        ),
        description,
    ]

    # By streaming the output, we allow the browser to render it as it is being
    # generated.
    async for part in self._plot_generator(content):
      # This is a bit ugly, but content_api.ProcessorPart.from_function_response
      # would render it as text instead of inline_data with
      # mime_type='text/html'.
      yield content_api.ProcessorPart(
          genai_types.Part(
              function_response=genai_types.FunctionResponse(
                  parts=[
                      genai_types.FunctionResponsePart(
                          inline_data=genai_types.FunctionResponseBlob(
                              data=part.text.encode(),
                              mime_type='text/html',
                          )
                      )
                  ],
              )
          ),
          # We use ui substream to send it directly to the UI.
          substream_name='ui',
      )


def create_dr_widget(
    api_key: str,
) -> processor.Processor:
  """Initializes the processor."""
  image_gen = ImageGenerator(api_key)
  plot_gen = PlotGenerator(api_key)
  turn_processor = genai_model.GenaiModel(
      api_key=api_key,
      model_name=MODEL,
      generate_content_config=genai_types.GenerateContentConfig(
          system_instruction=SYSTEM_INSTRUCTION,
          # We will be handling tool calls on the client side.
          automatic_function_calling=genai_types.AutomaticFunctionCallingConfig(
              disable=True
          ),
          tools=[
              image_gen.create_image_from_description,
              plot_gen.create_plot_from_description,
          ],
      ),
  )
  fc_processor = function_calling.FunctionCalling(
      model=realtime.LiveProcessor(turn_processor=turn_processor),
      fns=[
          image_gen.create_image_from_description,
          plot_gen.create_plot_from_description,
      ],
      is_bidi_model=True,
  )

  return fc_processor



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Switch processors to route parts to different processors."""

from __future__ import annotations

import asyncio
from collections.abc import AsyncIterable, Callable
from typing import Generic, Self, TypeAlias, TypeVar

from genai_processors import content_api
from genai_processors import processor
from genai_processors import streams

_T = TypeVar('_T')

ProcessorPart: TypeAlias = content_api.ProcessorPart
PartProcessor: TypeAlias = processor.PartProcessor
Processor: TypeAlias = processor.Processor


class Switch(Processor, Generic[_T]):
  """Switch between processors.

  Convenient way to create a processor that routes the parts of the input stream
  to different processors based on a condition (aka a case). The condition can
  be:

  1. a function that takes a `ProcessorPart` and returns a boolean. Example:
    ```python
    switch_processor = (
        switch.Switch()
        .case(content_api.is_audio, audio_processor)
        .case(content_api.is_video, video_processor)
        .default(processor.passthrough())
    )
    ```
  2. a function that takes any value returned by the match_fn passed in the
     constructor and returns a boolean. We have a shortcut for boolean functions
     that tests for equality, e.g. `lambda x: x == "a"`. They can be replaced
     with the value itself, e.g. `"a"`. Example:
    ```python
    # The match_fn is applied to the input part and the result is compared to
    # the value passed in the case() method.
    switch_processor = (
        switch.Switch(content_api.get_substream_name)
        .case("a", p)  # equivalent to .case(lambda x: x == "a")
        .case("b", q)  # equivalent to .case(lambda x: x == "b")
        .default(processor.passthrough())
    )
    ```

  The order of the parts in the output and input streams is only kept for parts
  returned by the same processor, i.e. two parts matching two different cases
  are not guaranteed to be in the same order in the input and output stream.

  PartProcessors can be used instead of Processor, they will be converted to
  Processor automatically. The `processor.passthrough()` (a PartProcessor) in
  the examples above is equivalent to passing
  `processor.passthrough().to_processor()`.

  If the cases involve `PartProcessor`s only, it is best to use the
  `PartSwitch` class, which is optimized for concurrent processing of parts.
  """

  def __init__(
      self,
      match_fn: Callable[[ProcessorPart], _T] | None = None,
  ):
    self._cases: list[tuple[Callable[[_T], bool], Processor]] = []
    self._match = match_fn
    self._default_set = False

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    input_queues = [asyncio.Queue() for _ in range(len(self._cases))]

    async def _triage():
      """Triage the input parts to the correct input queue."""
      async for part in content:
        for i, (filter_fn, _) in enumerate(self._cases):
          if filter_fn(part):
            await input_queues[i].put(part)
            break
      for q in input_queues:
        await q.put(None)

    triage_task = processor.create_task(_triage())

    # Process the parts in the input queues and merge all results.
    output_streams = [
        self._cases[i][1](streams.dequeue(queue))
        for i, queue in enumerate(input_queues)
    ]
    async for part in streams.merge(output_streams):
      yield part

    await triage_task

  def case(
      self,
      v: _T | Callable[[_T], bool],
      p: Processor | PartProcessor,
  ) -> Self:
    if self._default_set:
      raise ValueError(
          f'This case is added after the default processor is set: {v}'
      )
    if self._match is None:
      self._match = lambda x: x
    if isinstance(p, PartProcessor):
      case_processor = p.to_processor()
    else:
      case_processor = p
    if isinstance(v, Callable):
      self._cases.append((lambda x: v(self._match(x)), case_processor))
    else:
      self._cases.append((lambda x: v == self._match(x), case_processor))
    return self

  def default(self, p: Processor | PartProcessor) -> Self:
    if self._default_set:
      raise ValueError('The default processor is already set.')
    if isinstance(p, PartProcessor):
      self._cases.append((lambda x: True, p.to_processor()))
    else:
      self._cases.append((lambda x: True, p))
    self._default_set = True
    return self


class PartSwitch(PartProcessor, Generic[_T]):
  """Switch between part processors.

  Convenient way to create a switch processor that runs the first case that
  matches. A case is defined by a condition and a PartProcessor. The
  condition can be a function that takes a `ProcessorPart` and returns a boolean
  or a value. In the latter case, the condition is compared to the result of the
  match_fn passed to the constructor.

  By default, the switch processor does not return any part when no case
  matches. To return a part in that case, the default() method should be
  called after all cases have been added.

  Example usage:

  Simple match conditions based on equality:

  ```python
  # Applies content_api.as_text to the input part and checks equality with
  # the string "a" or "b".
  switch_processor = (
      switch.Switch(content_api.as_text)
      .case("a", p)
      .case("b", q)
      .default(processor.passthrough())
  ```

  Note that you can also compare parts directly by using the `lambda x: x`
  function in `Switch` but it is not recommended. For more complex match
  conditions, you can use cases operating on `part` directly as follows:

  ```python
  # Applies the lambda function defined in `case` on the input part to check
  # which case is valid.
  switch_processor = (
      switch.Switch()
      .case(lambda x: x.text.startswith("a"), p)
      .case(lambda x: x.text.startswith("b"), q)
      .default()
  )
  ```
  """

  def __init__(
      self,
      match_fn: Callable[[ProcessorPart], _T] | None = None,
  ):
    self._cases: list[tuple[Callable[[_T], bool], PartProcessor]] = []
    self._match = match_fn
    self._default_set = False

  async def call(
      self, part: ProcessorPart
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    for filter_fn, p in self._cases:
      if filter_fn(part):
        async for c in p(part):
          yield c
        break

  def case(
      self,
      v: _T | Callable[[_T], bool],
      p: PartProcessor,
  ) -> Self:
    if self._default_set:
      raise ValueError(
          f'This case is added after the default processor is set: {v}'
      )
    if self._match is None:
      self._match = lambda x: x
    if isinstance(v, Callable):
      self._cases.append((lambda x: v(self._match(x)), p))
    else:
      self._cases.append((lambda x: v == self._match(x), p))
    return self

  def default(self, p: PartProcessor) -> Self:
    if self._default_set:
      raise ValueError('The default processor is already set.')
    self._cases.append((lambda x: True, p))
    self._default_set = True
    return self



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Audio processors."""

import asyncio
from typing import AsyncIterable, Optional

from genai_processors import content_api
from genai_processors import processor
import pyaudio

ProcessorPart = content_api.ProcessorPart

# Audio output chunk size in bytes.
AUDIO_OUT_CHUNK_SIZE = 1024

# Add accepted audio formats here.
AudioFormats = pyaudio.paInt16 | pyaudio.paInt24


@processor.source()
async def PyAudioIn(
    pya: pyaudio.PyAudio,
    substream_name: str = 'realtime',
    audio_format: AudioFormats = pyaudio.paInt16,  # 16-bit PCM.
    channels: int = 1,
    rate: int = 24000,
    use_pcm_mimetype: bool = False,
):
  """Receives audio input and inserts it into the input stream.

  The audio input is received from the default input device.

  Args:
    pya: The pyaudio object to use for capturing audio.
    substream_name: The name of the substream that will contain all the audio
      parts captured from the mic.
    audio_format: The audio format to use for the audio.
    channels: The number of channels in the audio.
    rate: The sample rate of the audio.
    use_pcm_mimetype: Whether to use PCM mimetype instead of the more specific
      l16 or l24 mimetype.
  """
  mimetype = 'audio/'
  match audio_format:
    case pyaudio.paInt16:
      mimetype += 'pcm' if use_pcm_mimetype else 'l16'
    case pyaudio.paInt24:
      mimetype += 'pcm' if use_pcm_mimetype else 'l24'
    case _:
      raise ValueError(f'Unsupported audio format: {format}')
  mimetype = f'{mimetype};rate={rate}'

  mic_info = pya.get_default_input_device_info()
  audio_stream = await asyncio.to_thread(
      pya.open,
      format=audio_format,
      channels=channels,
      rate=rate,
      input=True,
      input_device_index=mic_info['index'],
      frames_per_buffer=AUDIO_OUT_CHUNK_SIZE,
  )
  if __debug__:  # pylint: disable=undefined-variable
    kwargs = {'exception_on_overflow': False}
  else:
    kwargs = {}

  while True:
    data = await asyncio.to_thread(
        audio_stream.read, AUDIO_OUT_CHUNK_SIZE, **kwargs
    )
    yield ProcessorPart(
        data, mimetype=mimetype, substream_name=substream_name, role='user'
    )


class PyAudioOut(processor.Processor):
  """Receives audio output from a live session and talks back to the user.

  Uses pyaudio to play audio back to the user.

  All non audio parts are passed through based on the `passthrough_audio` param
  passed to the constructor.

  Combine this processor with `RateLimitAudio` to receive the audio chunks at
  the time where they need to be played back to the user.
  """

  def __init__(
      self,
      pya: pyaudio.PyAudio,
      audio_format=pyaudio.paInt16,  # 16-bit PCM.
      channels: int = 1,
      rate: int = 24000,
      passthrough_audio: bool = False,
  ):
    self._pya = pya
    self._format = audio_format
    self._channels = channels
    self._rate = rate
    self._passthrough_audio = passthrough_audio

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:
    """Receives audio output from a live session."""
    audio_output = asyncio.Queue[Optional[ProcessorPart]]()

    stream = await asyncio.to_thread(
        self._pya.open,
        format=self._format,
        channels=self._channels,
        rate=self._rate,
        output=True,
    )

    async def play_audio():  # pylint: disable=invalid-name
      while part := await audio_output.get():
        if part.part.inline_data is not None:
          await asyncio.to_thread(stream.write, part.part.inline_data.data)

    play_audio_task = processor.create_task(play_audio())

    async for part in content:
      if content_api.is_audio(part.mimetype):
        audio_output.put_nowait(part)
        if self._passthrough_audio:
          yield part
      else:
        yield part
    await audio_output.put(None)
    play_audio_task.cancel()



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities to add more debug information to content streams."""

import asyncio
from collections.abc import AsyncIterable
import time

from absl import logging
from genai_processors import content_api
from genai_processors import processor

ProcessorPart = content_api.ProcessorPart


class TTFTSingleStream(processor.Processor):
  """Wraps a processor to provide performance messaging."""

  def __init__(self, message: str, p: processor.Processor):
    """Wraps a processor to provide performance messaging.

      Should only be used for processors that consume their entire input before
      producing output (such as non-streaming or unidirectional/single streaming
      model calls). The TTFT is estimated by waiting first that the inputs
      stream is
      completely sent to the processor (`start` time is then set). When the
      processor outputs its first token, the duration from `start` is then
      reported.

    In a bidirectional streaming setup, the TTFT will not be reported at all.

    Args:
      message: header of the status chunk that will be returned. It is used to
        identify different calls to this function.
      p: processor for which we need to compute ttft. self._message = message
        self._p = p self._start = None self._ttft = None self._model_call_event
        = asyncio.Event() self._model_call_event.clear()
    """
    self._message = message
    self._p = p
    self._start = None
    self._ttft = None
    self._model_call_event = asyncio.Event()
    self._model_call_event.clear()

  def model_call_event(self) -> asyncio.Event:
    """Returns an event that is set when the wrapped processor has all parts.

    The event is set when the wrapped processor has all the input parts and
    is about to start generating the output.

    The event starts in a cleared state when the first part of the input
    stream is yielded. It is also cleared at the end of the wrapped processor,
    when all the output parts have been yielded.

    Its default value is unset and this event is set only for a short time
    during the call.

    Returns:
      An event that is set when the model call is started, that is when all the
      input parts have been sent to the wrapped processor.
    """
    return self._model_call_event

  def _ttft_processor(self) -> processor.Processor:

    @processor.processor_function
    async def log_on_close(
        content: AsyncIterable[ProcessorPart],
    ) -> AsyncIterable[ProcessorPart]:
      self._model_call_event.clear()
      async for part in content:
        yield part
      self._start = time.perf_counter()
      self._model_call_event.set()

    @processor.processor_function
    async def log_on_first(
        content: AsyncIterable[ProcessorPart],
    ) -> AsyncIterable[ProcessorPart]:
      first_part = True
      async for part in content:
        if first_part:
          if self._start is not None:
            # We have finished going over all parts in the input content.
            # We can compute a proper TTFT.
            duration = time.perf_counter() - self._start
            self._ttft = duration
            self._message += f' TTFT={duration:.2f} seconds'
          else:
            # The input content is still being sent, typical of a bidi setup.
            # TTFT is 0 here by convention. We can't compute it yet.
            self._message += ' TTFT=0 - input stream still open.'
          yield processor.status(ProcessorPart(self._message))
        first_part = False
        yield part

    return log_on_close + self._p + log_on_first

  def ttft(self) -> float | None:
    """Returns the TTFT of the wrapped processor.

    Returns:
      the TTFT of the wrapped processor or None if the processor has not been
      called yet.
    """
    return self._ttft

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:
    async for chunk in self._ttft_processor()(content):
      yield chunk


def debug_string(part: ProcessorPart) -> str:
  return f'role {part.role} substream {part.substream_name} {part!r}'


def log_stream(message: str) -> processor.Processor:
  """Return a function that logs every part of a stream."""

  @processor.processor_function
  async def p(
      content: AsyncIterable[ProcessorPart],
  ) -> AsyncIterable[ProcessorPart]:
    async for part in content:
      logging.info('%s: %s', message, debug_string(part))
      yield part
    logging.info('%s: done', message)

  return p


def print_stream(message: str) -> processor.Processor:
  """Return a function that prints every part of a stream."""

  @processor.processor_function
  async def p(
      content: AsyncIterable[ProcessorPart],
  ) -> AsyncIterable[ProcessorPart]:
    async for part in content:
      print(f'{message}: {debug_string(part)}')
      yield part

  return p


def log_queue(
    message: str, queue: asyncio.Queue[ProcessorPart | None]
) -> asyncio.Queue[ProcessorPart | None]:
  """Return a function that logs every part of a queue."""
  output_queue = asyncio.Queue()

  async def log_and_output():
    while (part := await queue.get()) is not None:
      queue.task_done()
      logging.info('%s: %s', message, debug_string(part))
      output_queue.put_nowait(part)
    output_queue.put_nowait(None)

  processor.create_task(log_and_output())
  return output_queue



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Streams audio content parts from text parts.

Uses Google Text-To-Speech API to generate audio parts from text parts.

Install google cloud Text-To-Speech speech client with:

```python
pip install --upgrade google-cloud-texttospeech
```

See the `text_to_speech_cli.py` script for a usage example and how to test it
locally.
"""

import asyncio
from collections.abc import AsyncIterable

from genai_processors import content_api
from genai_processors import processor
from genai_processors import streams
from google.cloud import texttospeech_v1 as texttospeech

ProcessorPart = content_api.ProcessorPart


class TextToSpeech(processor.Processor):
  """Streams audio content parts from text parts."""

  def __init__(
      self,
      project_id: str,
      language_code: str = 'en-US',
      voice_name: str = 'en-US-Chirp3-HD-Charon',
      with_text_passthrough: bool = True,
  ):
    """Initializes the TextToSpeech processor.

    This processor uses the Google Text-To-Speech API to generate audio parts
    from text parts. The audio parts are yielded as `ProcessorPart` objects with
    the `audio_content` attribute set to the audio bytes and the `mimetype`
    attribute set to `audio/l16;rate=24000`.

    Args:
      project_id: The project ID to use for the Text-To-Speech API.
      language_code: The language code to use for the Text-To-Speech API.
      voice_name: The voice name to use for the Text-To-Speech API. See list of
        voices here: https://cloud.google.com/text-to-speech/docs/chirp3-hd
      with_text_passthrough: Whether to passthrough the text parts to the output
        stream. When set to True, the text parts are yielded back.
    """
    streaming_config = texttospeech.StreamingSynthesizeConfig(
        streaming_audio_config=texttospeech.StreamingAudioConfig(
            audio_encoding=texttospeech.AudioEncoding.PCM,
            sample_rate_hertz=24000,
        ),
        voice=texttospeech.VoiceSelectionParams(
            name=voice_name,
            language_code=language_code,
        ),
    )
    self._config_request = texttospeech.StreamingSynthesizeRequest(
        streaming_config=streaming_config
    )
    self._project_id = project_id
    self._with_text_passthrough = with_text_passthrough

  async def call(
      self, content: AsyncIterable[ProcessorPart]
  ) -> AsyncIterable[ProcessorPart]:
    """Streams audio content parts from text parts.

    The order between TTS-processed parts and pass-through parts is not
    maintained. This processor treats all its inputs as realtime and sends them
    to the output as soon as possible. Order within TTS-processed an
    pass-through parts is maintained, but we don't wait for the TTS result to
    emit the next pass-through part.

    Args:
      content: The input stream of content to process. Non-text parts are passed
        through unchanged.

    Yields:
      The audio parts generated by the Text-To-Speech API from the text parts in
      `content`. If `with_text_passthrough` is True, the text parts are yielded
      back as well. All non-text parts are yielded unchanged.
    """
    # The output queue is used to yield the audio parts unchanged in the output
    # stream when _with_text_passthrough is True.
    output_queue = asyncio.Queue[ProcessorPart | None]()
    first_chunk_received = asyncio.Event()

    async def request_stream(
        request_queue: asyncio.Queue[
            texttospeech.StreamingSynthesizeRequest | None
        ],
    ):
      try:
        request_queue.put_nowait(self._config_request)
        async for part in content:
          if (
              not content_api.is_text(part.mimetype)
              or self._with_text_passthrough
          ):
            output_queue.put_nowait(part)
          if not content_api.is_text(part.mimetype) or not part.text:
            continue
          first_chunk_received.set()
          request_queue.put_nowait(
              texttospeech.StreamingSynthesizeRequest(
                  input=texttospeech.StreamingSynthesisInput(
                      text=part.text,
                  )
              )
          )
      finally:
        first_chunk_received.set()
        request_queue.put_nowait(None)

    async def send_text_to_speech_requests():
      try:
        request_queue = asyncio.Queue[
            texttospeech.StreamingSynthesizeRequest | None
        ]()
        enqueue_request_task = processor.create_task(
            request_stream(request_queue)
        )
        # Wait until the first request is sent to the Speech API to avoid the
        # client being created before the first request is sent.
        # The client can indeed only stay up for 5 seconds without any request
        # and then it needs to be re-created.
        await first_chunk_received.wait()
        client = texttospeech.TextToSpeechAsyncClient()
        streaming_responses = await client.streaming_synthesize(
            requests=streams.dequeue(request_queue)
        )

        async for response in streaming_responses:
          output_queue.put_nowait(
              ProcessorPart(
                  response.audio_content,
                  mimetype='audio/l16;rate=24000',
                  role='model',
              )
          )

        await enqueue_request_task
      finally:
        output_queue.put_nowait(None)

    send_task = processor.create_task(send_text_to_speech_requests())
    while part := await output_queue.get():
      yield part
    await send_task



# Copyright 2026 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Audio processors."""

from collections.abc import AsyncIterable
import io
import re
import wave

from genai_processors import content_api
from genai_processors import processor

_DEFAULT_SAMPLE_RATE = 24000


def _l16_to_wav(
    l16_audio_data: bytes, sample_rate: int, num_channels: int = 1
) -> bytes:
  """Converts L16 audio data to the WAV file format.

  Args:
      l16_audio_data: The raw L16 audio data.
      sample_rate: The sample rate of the audio (e.g., 8000, 16000).
      num_channels: The number of audio channels (e.g., 1 for mono, 2 for
        stereo).

  Returns:
      bytes: A byte string containing the WAV file data.
  """
  wav_stream = io.BytesIO()

  with wave.open(wav_stream, 'wb') as wav_file:
    wav_file.setnchannels(num_channels)
    wav_file.setsampwidth(2)  # L16 is 16-bit, which is 2 bytes
    wav_file.setframerate(sample_rate)
    wav_file.writeframes(l16_audio_data)

  return wav_stream.getvalue()


class AudioToWav(processor.Processor):
  """Concatenates consecutive audio parts into a single WAV part.

  Non live Gemini models don't handle lots of small audio parts well. We have to
  concatenate them into a single part to make sure they are handled correctly.
  """

  def __init__(self):
    self._buffer: list[bytes] = []
    self._mimetype: str | None = None

  async def _flush(
      self,
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    if self._buffer:
      sample_rate = _DEFAULT_SAMPLE_RATE
      if self._mimetype:
        match = re.search(r'rate=(\d+)', self._mimetype)
        if match:
          sample_rate = int(match.group(1))

      part = content_api.ProcessorPart(
          # Non Live Gemini API models do not support audio/l16 or audio/pcm
          # mime types. So we have to convert it to audio.wav.
          _l16_to_wav(b''.join(self._buffer), sample_rate),
          mimetype='audio/wav',
      )
      self._buffer = []
      self._mimetype = None
      yield part

  async def call(
      self,
      content: AsyncIterable[content_api.ProcessorPartTypes],
  ) -> AsyncIterable[content_api.ProcessorPartTypes]:
    """Concatenates consecutive audio parts."""
    async for part in content:
      if content_api.is_audio(part.mimetype):
        if not (
            part.mimetype.startswith('audio/l16')
            or part.mimetype.startswith('audio/pcm')
        ):
          raise ValueError(
              'Only audio/l16 or audio/pcm is supported. Unsupported audio'
              f' mimetype: {part.mimetype}'
          )

        if self._buffer and part.mimetype != self._mimetype:
          async for flushed_part in self._flush():
            yield flushed_part

        self._buffer.append(part.bytes)
        self._mimetype = part.mimetype
      else:
        async for flushed_part in self._flush():
          yield flushed_part
        yield part

    async for flushed_part in self._flush():
      yield flushed_part



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

r"""Live WebSocket server for AI Studio.

This server takes a GenAI processor factory as input and wrap it around a
WebSocket server. This allows any GenAI processor to be used with external
applications such as AI Studio.

The server expects a stringified json dictionary of content_api.ProcessorPart as
input from the client. The bytes field should be base64 encoded. The mime_type
field should be set to the correct mime_type. The substream_name field should be
set to 'realtime' for images and audio streamed live from the UI.

Empty ProcessorPart with the following metadata fields can also be sent:

-  `'reset' : True` - This is sent when to reset the server.
-  `'mic: 'off'` - This is sent when the client has turned off the mic.

The server returns a stringified json dictionary of content_api.ProcessorPart as
output. The server can also send the following empty ProcessorPart of mime type
`application/x-state` with metadata:

-  `'generation_complete' : True` - This is sent when the processor has
    finished generating a response.
-  `'interrupted' : True` - This is sent when the processor has been
    interrupted, e.g. due to a `start of speech` signal.
-  `'health_check' : True` - This is sent periodically to check if the
    connection is alive.


Example of messages going back and forth between the client and the server:

## From the client
{
  "part": {
    "text": "Hello World",
    }
  "role": "user",
  "substream_name": "realtime",
  }
}

## From the client
{
  "part": {
    "inline_data": {
      "data": "SGVsbG8gV29ybGQ=", # Base64 encoded
      "mime_type": "audio/webm",
    }
  "role": "user",
  "metadata": {
    "audio_stream_end": true,
    }
  }
}

## From the server
{
  "part": {
    "text": "Hello World",
    }
  "role": "model",
  "metadata": {
    "generation_complete": true,
    }
  }
}

## From the server
{
  "metadata": {
    "health_check": true,
    }
  }
  "mimetype": "application/x-state",
}

__WARNING__: this is a work in progress and is provided here for convenience and
for prototyping quickly. We will likely make backwards incompatible changes to
comply with a more stable GenAI API between clients and servers.
"""

import asyncio
import base64
import functools
import json
import os
import time
import traceback
from typing import Any, AsyncIterable, Callable

from absl import logging
from genai_processors import content_api
from genai_processors import processor
from genai_processors.dev import trace_file
import pydantic
from websockets.asyncio.server import serve
from websockets.asyncio.server import ServerConnection
from websockets.exceptions import ConnectionClosed

# Mimetype for a command part. A command represents a specific instruction for
# the server to trigger actions or modify its state.
_COMMAND_MIMETYPE = 'application/x-command'

# Config parts can be sent to this server to configure the server.
_CONFIG_MIMETYPE = 'application/x-config'

# Mimetype to represent the state of either the client or the server.
_STATE_MIMETYPE = 'application/x-state'


def is_reset_command(part: content_api.ProcessorPart) -> bool:
  """Returns whether the part is a reset command."""
  return (
      part.mimetype == _COMMAND_MIMETYPE
      and part.get_metadata('command', None) == 'reset'
  )


def is_config(part: content_api.ProcessorPart) -> bool:
  """Returns whether the part is a config."""
  return part.mimetype == _CONFIG_MIMETYPE


def is_mic_off(part: content_api.ProcessorPart) -> bool:
  """Returns whether the part indicates the client has turned off the mic."""
  return (
      part.mimetype == _STATE_MIMETYPE
      and part.get_metadata('mic', None) == 'off'
  )


pydantic_converter = pydantic.TypeAdapter(Any)


def clean_encoder(o: Any) -> Any:
  """A custom encoder for json.dumps that handles bytes with base64 encoding.

  The from_dict(mode='json') in ProcessorPart does not encode bytes in utf-8
  mode and this causes issues when sending ProcessorPart to JS clients (wrong
  padding, etc.). This function is used to handle bytes to base64 encoding to
  match the behaviour of the JS side. It can be used as an argument for
  json.dumps. Use it with JS dictionaries obtained from the ProcessorPart
  to_dict(mode='python') method (leaving bytes as is).

  Args:
    o: The object to be encoded. Uses pydantic to convert to json excpet when it
      is bytes. Then uses base64 encoding for bytes followed by utf-8 decoding.

  Returns:
    The object encoded as a string.
  """
  if isinstance(o, bytes):
    # Convert to Standard Base64 (happy path for JS atob)
    return base64.b64encode(o).decode('utf-8')
  return pydantic_converter.dump_python(o, mode='json')


class AIStudioConnection:
  """A WebSocket connection with AI Studio."""

  def __init__(self, ais_ws: ServerConnection):
    self._ais_ws = ais_ws
    self.is_resetting = False
    # Config dictionary directly transferred to the genai processor.
    self.processor_config = {}

  async def send(
      self,
      output_stream: AsyncIterable[content_api.ProcessorPart],
  ):
    """Sends audio to AIS."""
    async for part in output_stream:
      if self.is_resetting:
        return
      if part.get_metadata('generation_complete', False) or part.get_metadata(
          'interrupted', False
      ):
        part = content_api.ProcessorPart(
            '',
            mimetype=_STATE_MIMETYPE,
            metadata=part.metadata,
        )
        await self._ais_ws.send(json.dumps(part.to_dict()))
      else:
        await self._ais_ws.send(
            json.dumps(part.to_dict(mode='python'), default=clean_encoder)
        )

  async def receive(self) -> AsyncIterable[content_api.ProcessorPart]:
    """Reads chunks from AIS."""
    async for part_dict in self._ais_ws:
      json_dict = json.loads(part_dict)
      if 'part' not in json_dict:
        json_dict['part'] = {'text': ''}
      part = content_api.ProcessorPart.from_dict(
          data=json_dict,
      )
      if content_api.is_image(part.mimetype) or content_api.is_audio(
          part.mimetype
      ):
        part.substream_name = 'realtime'
        part.role = 'user'
        yield part
      elif content_api.is_text(part.mimetype):
        part.role = 'user'
        part.metadata['turn_complete'] = True
        yield part
      elif is_mic_off(part):
        yield content_api.ProcessorPart(
            '',
            substream_name='realtime',
            role='user',
            metadata={'audio_stream_end': True},
        )
      elif is_reset_command(part):
        # Stop reading from the WebSocket until the processor has been reset.
        logging.debug(
            "%s - RESET command received. Resetting the processor's state.",
            time.perf_counter(),
        )
        self.is_resetting = True
        return
      elif is_config(part) and part.metadata:
        self.processor_config = part.metadata
        logging.info(
            '%s - Config received: %s',
            time.perf_counter(),
            self.processor_config,
        )
        self.is_resetting = True
        return
      else:
        logging.warning('Unknown input part type: %s', part.mimetype)


async def live_server(
    processor_factory: Callable[[dict[str, Any]], processor.Processor],
    trace_dir: str | None,
    max_size_bytes: int | None,
    ais_websocket: ServerConnection,
):
  """Runs the processor on AI Studio input/output streams."""
  ais = AIStudioConnection(ais_websocket)

  # Running in a loop as the agent can receive a RESET command from AIS, in
  # which case the live loop needs to be reinitialized.
  while True:
    try:
      logging.info('Starting new server session...')
      live_processor = processor_factory(ais.processor_config)
      if trace_dir:
        async with trace_file.SyncFileTrace(
            trace_dir=trace_dir,
            name='live_server',
            max_size_bytes=max_size_bytes,
        ):
          await ais.send(live_processor(ais.receive()))
      else:
        await ais.send(live_processor(ais.receive()))
    except Exception as e:  # pylint: disable=broad-exception-caught
      logging.info(
          '%s - Resetting live server after receiving error :'
          ' %s\n\nTraceback:\n%s',
          time.perf_counter(),
          e,
          traceback.format_exc(),
      )

    ais.is_resetting = False

    # Exit the loop if the connection is closed.
    try:
      await ais_websocket.send(
          json.dumps(
              content_api.ProcessorPart(
                  '', metadata={'health_check': True}
              ).to_dict()
          )
      )
    except ConnectionClosed:
      logging.debug('Connection between AIS and agent has been closed.')
      break


async def run_server(
    processor_factory: Callable[[dict[str, Any]], processor.Processor],
    port: int = 8765,
    trace_dir: str | None = None,
    max_size_bytes: int | None = None,
) -> None:
  """Starts the WebSocket server."""
  if trace_dir:
   os.makedirs(trace_dir, exist_ok=True)

  # set this future to exit the server
  stop = asyncio.get_running_loop().create_future()

  async with serve(
      handler=functools.partial(
          live_server,
          processor_factory,
          trace_dir,
          max_size_bytes,
      ),
      host='localhost',
      port=port,
      max_size=2 * 1024 * 1024,  # 2 MiB
  ):
    print(f'Server started on port {port}')
    await stop



# Copyright 2025 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Context vars for managing task groups."""

import asyncio
from collections.abc import Coroutine, Iterable
import contextvars
from typing import Any, TypeVar

from absl import logging


_PROCESSOR_TASK_GROUP: contextvars.ContextVar['CancellableContextTaskGroup'] = (
    contextvars.ContextVar('processor_task_group')
)

PROMPT_STREAM = 'prompt'
# Processors can yield debug information in this substream thus embedding it in
# the conversation timeline.
DEBUG_STREAM = 'debug'
# When performing long tasks, processors can send status updates e.g. 1-line
# summary of what they are doing.
STATUS_STREAM = 'status'
# Allows processors and in particular tools to send Parts directly to the user
# instead of passing them through the model.
UI_STREAM = 'ui'

# Reserved substreams (and any substream with a reserved substream as a prefix)
# are not passed to the processor but are instead "captured" and yielded
# immediately to the output. This is useful for out-of-band data that should not
# be processed by the model or other processors in the chain (e.g. debug logs,
# status updates, UI events).
#
# Processors can add their own reserved substreams using the `context`
# context manager.
_PROCESSOR_RESERVED_SUBSTREAMS: contextvars.ContextVar[frozenset[str]] = (
    contextvars.ContextVar(
        'processor_reserved_substreams',
        default=frozenset({DEBUG_STREAM, STATUS_STREAM, UI_STREAM}),
    )
)


def raise_flattened_exception_group(exception: Exception):
  e = exception
  while isinstance(e, ExceptionGroup):
    e = e.exceptions[0]
  if e is exception:
    raise exception
  else:
    raise e from exception


class CancellableContextTaskGroup(asyncio.TaskGroup):
  """TaskGroup that adds itself to a contextvar to be accessed by create_task.

  Includes a method for cancelling all tasks in the group.
  """

  def __init__(
      self, *args, reserved_substreams: Iterable[str] | None = None, **kwargs
  ):
    super().__init__(*args, **kwargs)
    self._cancel_tasks = set()
    self._reserved_substreams = reserved_substreams

  def create_task(self, *args, **kwargs) -> asyncio.Task:
    t = super().create_task(*args, **kwargs)
    self._cancel_tasks.add(t)
    t.add_done_callback(self._cancel_tasks.discard)
    return t

  async def __aenter__(self) -> 'CancellableContextTaskGroup':
    self._current_taskgroup_token = _PROCESSOR_TASK_GROUP.set(self)
    if self._reserved_substreams is not None:
      self._reserved_substreams_token = _PROCESSOR_RESERVED_SUBSTREAMS.set(
          frozenset(self._reserved_substreams)
      )
    else:
      self._reserved_substreams_token = None
    return await super().__aenter__()

  async def __aexit__(self, et, exc, tb):
    try:
      return await super().__aexit__(et, exc, tb)
    except BaseExceptionGroup as e:
      raise_flattened_exception_group(e)
    finally:
      try:
        _PROCESSOR_TASK_GROUP.reset(self._current_taskgroup_token)
        if self._reserved_substreams_token is not None:
          _PROCESSOR_RESERVED_SUBSTREAMS.reset(self._reserved_substreams_token)
      except ValueError:
        # ValueError is raised when the Context self._current_taskgroup_token
        # was created in doesn't match the current Context.
        # This can happen when an async generator is garbage collected.
        # The task loop will call loop.call_soon(loop.create_task, agen.aclose).
        # aclose raises GeneratorExit in the async generator which is seen here
        # but from a different context.
        if et is GeneratorExit:
          logging.log_first_n(
              logging.WARNING,
              'GeneratorExit was seen in processors.context. This'
              ' indicates that the async generator that opened the context has'
              ' been closed from a different context. For example, it has been'
              ' garbage collected. This usually means a task executing the'
              ' generator was also garbage collected. Consider turning on'
              ' asyncio debug mode to investigate further.',
              1,
          )
          pass

  def cancel(self):
    for task in self._cancel_tasks:
      task.cancel()


def context(
    reserved_substreams: Iterable[str] | None = None,
) -> CancellableContextTaskGroup:
  return CancellableContextTaskGroup(reserved_substreams=reserved_substreams)


def task_group() -> CancellableContextTaskGroup | None:
  return _PROCESSOR_TASK_GROUP.get(None)


def get_reserved_substreams() -> frozenset[str]:
  return _PROCESSOR_RESERVED_SUBSTREAMS.get()


def is_reserved_substream(substream_name: str) -> bool:
  return any(
      substream_name.startswith(prefix) for prefix in get_reserved_substreams()
  )


# If a task is created without a task group then a reference to it must be kept.
_without_context_background_tasks = set()


def create_task(*args, **kwargs) -> asyncio.Task:
  """Creates a task that uses the context TaskGroup.

  If no context is available then `asyncio.create_task` will be used.

  Args:
    *args: Positional arguments to pass to `asyncio.create_task`.
    **kwargs: Keyword arguments to pass to `asyncio.create_task`.

  Returns:
    An asyncio task.
  """
  tg = task_group()

  if tg is None:
    task = asyncio.create_task(*args, **kwargs)
    _without_context_background_tasks.add(task)
    task.add_done_callback(_without_context_background_tasks.discard)
    return task

  return tg.create_task(*args, **kwargs)


_T = TypeVar('_T')


async def context_cancel_coro(
    f: Coroutine[Any, Any, _T],
) -> _T:
  """Wrapper that cancels all tasks in context if the wrapper is cancelled."""
  async with context() as ctx:
    try:
      return await f
    except asyncio.CancelledError:
      ctx.cancel()
      raise
